from __future__ import annotations

import asyncio
from collections import defaultdict
from collections.abc import Awaitable, Callable, Iterator
from contextlib import contextmanager
from datetime import datetime
from functools import partial
import importlib.resources
import os
from pathlib import Path
import typing
from typing import Any, TypeVar

import aiohttp
import aiohttp.web
from aiohttp_rpc import JsonRpcMethod
from aiohttp_rpc import middlewares as rpc_middlewares
from aiohttp_rpc.errors import InvalidParams as InvalidParamsError
from aiohttp_rpc.errors import ServerError
from aiohttp_rpc.server import WsJsonRpcServer
import click
from loguru import logger
from pydantic import BaseModel, ValidationError
import sqlalchemy as sa
import toga
from typing_extensions import Concatenate, Literal, ParamSpec, TypeAlias, TypedDict
from yarl import URL

from instawow import __version__, db, matchers, models
from instawow import results as R
from instawow.cataloguer import CatalogueEntry
from instawow.common import Flavour, Strategy
from instawow.config import Config, GlobalConfig
from instawow.manager import LocksType, Manager, TraceRequestCtx, init_web_client, is_outdated
from instawow.resolvers import Defn
from instawow.utils import evolve_model_obj
from instawow.utils import run_in_thread as t
from instawow.utils import uniq

from . import frontend

_T = TypeVar('_T')
_P = ParamSpec('_P')
ManagerBoundCoroFn: TypeAlias = 'Callable[Concatenate[Manager, _P], Awaitable[_T]]'


LOCALHOST = '127.0.0.1'


class _ConfigError(ServerError):
    code = -32001
    message = 'invalid configuration parameters'


@contextmanager
def _reraise_validation_error(
    error_class: type[ServerError | InvalidParamsError] = ServerError,
    values: dict[Any, Any] | None = None,
) -> Iterator[None]:
    try:
        yield
    except ValidationError as error:
        errors = error.errors()
        logger.info(f'invalid request: {(values, errors)}')
        raise error_class(data=errors) from error


@t
def _read_config(profile: str) -> Config:
    with _reraise_validation_error(_ConfigError):
        return Config.read(None, profile)


methods: list[tuple[str, type[BaseParams]]] = []


def _register_method(method: str):
    def inner(param_class: type[BaseParams]):
        methods.append((method, param_class))
        return param_class

    return inner


class BaseParams(BaseModel):
    @classmethod
    def bind(
        cls, method: str, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> JsonRpcMethod:
        async def respond(**kwargs: Any) -> BaseModel:
            with _reraise_validation_error(InvalidParamsError, kwargs):
                params = cls.parse_obj(kwargs)
            return await params.respond(managers, app_window)

        return JsonRpcMethod(respond, name=method)

    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> Any:
        raise NotImplementedError


class _ProfileParamMixin(BaseModel):
    profile: str


class _DefnParamMixin(BaseModel):
    defns: typing.List[Defn]


@_register_method('config/write')
class WriteConfigParams(_ProfileParamMixin, BaseParams):
    addon_dir: Path
    game_flavour: Flavour
    infer_game_flavour: bool

    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> Config:
        async with managers.locks['modify profile', self.profile]:
            with _reraise_validation_error(_ConfigError):
                config = Config(
                    global_config=await t(GlobalConfig.read)(),
                    profile=self.profile,
                    addon_dir=self.addon_dir,
                    game_flavour=self.game_flavour,
                )
                if self.infer_game_flavour:
                    values = {
                        **config.dict(),
                        'game_flavour': Config.infer_flavour(config.addon_dir),
                    }
                    config = Config(**values)

                await t(config.write)()

            # Unload the ``Manager`` instance corresponding to this profile for
            # the config to be reloaded on the next request
            managers.unload(config.profile)

            return config


@_register_method('config/read')
class ReadConfigParams(_ProfileParamMixin, BaseParams):
    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> Config:
        return await _read_config(self.profile)


@_register_method('config/delete')
class DeleteConfigParams(_ProfileParamMixin, BaseParams):
    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> None:
        async def delete_profile(manager: Manager):
            await t(manager.config.delete)()
            managers.unload(self.profile)

        await managers.run(self.profile, delete_profile)


@_register_method('config/list')
class ListProfilesParams(BaseParams):
    @t
    def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> list[str]:
        return GlobalConfig().list_profiles()


class Source(TypedDict):
    source: str
    name: str
    supported_strategies: list[Strategy]
    supports_rollback: bool
    changelog_format: str


@_register_method('sources/list')
class ListSourcesParams(_ProfileParamMixin, BaseParams):
    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> list[Source]:
        manager = await managers.get_manager(self.profile)
        return [
            {
                'source': r.source,
                'name': r.name,
                'supported_strategies': sorted(r.strategies, key=list(Strategy).index),
                'supports_rollback': Strategy.version in r.strategies,
                'changelog_format': r.changelog_format.value,
            }
            for r in manager.resolvers.values()
        ]


@_register_method('list')
class ListInstalledParams(_ProfileParamMixin, BaseParams):
    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> list[models.Pkg]:
        manager = await managers.get_manager(self.profile)
        installed_pkgs = (
            manager.database.execute(sa.select(db.pkg).order_by(sa.func.lower(db.pkg.c.name)))
            .mappings()
            .all()
        )
        return [models.Pkg.from_row_mapping(manager.database, p) for p in installed_pkgs]


@_register_method('search')
class SearchParams(_ProfileParamMixin, BaseParams):
    search_terms: str
    limit: int
    sources: typing.Set[str]
    start_date: typing.Optional[datetime]
    installed_only: bool

    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> list[CatalogueEntry]:
        return await managers.run(
            self.profile,
            partial(
                Manager.search,
                search_terms=self.search_terms,
                limit=self.limit,
                sources=self.sources,
                start_date=self.start_date,
                installed_only=self.installed_only,
            ),
        )


class SuccessResult(TypedDict):
    status: Literal['success']
    addon: models.Pkg


class ErrorResult(TypedDict):
    status: Literal['failure', 'error']
    message: str


@_register_method('resolve')
class ResolveParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    async def _resolve(self, manager: Manager) -> dict[Defn, Any]:
        def extract_source(defn: Defn):
            if defn.source == '*':
                pair = manager.pair_uri(defn.alias)
                if pair:
                    source, alias = pair
                    defn = evolve_model_obj(defn, source=source, alias=alias)
            return defn

        return await manager.resolve(list(map(extract_source, self.defns)))

    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> list[SuccessResult | ErrorResult]:
        results = await managers.run(self.profile, self._resolve)
        return [
            {'status': 'success', 'addon': r}
            if isinstance(r, models.Pkg)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('install')
class InstallParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    replace: bool

    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> list[SuccessResult | ErrorResult]:
        results = await managers.run(
            self.profile, partial(Manager.install, defns=self.defns, replace=self.replace)
        )
        return [
            {'status': r.status, 'addon': r.pkg}
            if isinstance(r, R.PkgInstalled)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('update')
class UpdateParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> list[SuccessResult | ErrorResult]:
        results = await managers.run(
            self.profile, partial(Manager.update, defns=self.defns, retain_defn_strategy=True)
        )
        return [
            {'status': r.status, 'addon': r.new_pkg}
            if isinstance(r, R.PkgUpdated)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('remove')
class RemoveParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    keep_folders: bool

    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> list[SuccessResult | ErrorResult]:
        results = await managers.run(
            self.profile, partial(Manager.remove, defns=self.defns, keep_folders=self.keep_folders)
        )
        return [
            {'status': r.status, 'addon': r.old_pkg}
            if isinstance(r, R.PkgRemoved)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('pin')
class PinParams(_ProfileParamMixin, _DefnParamMixin, BaseParams):
    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> list[SuccessResult | ErrorResult]:
        results = await managers.run(self.profile, partial(Manager.pin, defns=self.defns))
        return [
            {'status': r.status, 'addon': r.pkg}
            if isinstance(r, R.PkgInstalled)
            else {'status': r.status, 'message': r.message}
            for r in results.values()
        ]


@_register_method('get_changelog')
class GetChangelogParams(_ProfileParamMixin, BaseParams):
    changelog_url: str

    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> str:
        return await managers.run(
            self.profile, partial(Manager.get_changelog, uri=self.changelog_url)
        )


class AddonMatch(TypedDict):
    folders: list[AddonMatch_AddonFolder]
    matches: list[models.Pkg]


class AddonMatch_AddonFolder(TypedDict):
    name: str
    version: str


@_register_method('reconcile')
class ReconcileParams(_ProfileParamMixin, BaseParams):
    matcher: Literal['toc_source_ids', 'folder_name_subsets', 'addon_names_with_folder_names']

    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> list[AddonMatch]:
        leftovers = await managers.run(self.profile, t(matchers.get_unreconciled_folders))
        match_groups: matchers.FolderAndDefnPairs = await managers.run(
            self.profile, partial(getattr(matchers, f'match_{self.matcher}'), leftovers=leftovers)
        )
        resolve_results = await managers.run(
            self.profile,
            partial(Manager.resolve, defns=uniq(d for _, b in match_groups for d in b)),
        )
        matches = [
            (a, m)
            for a, s in match_groups
            for m in ([r for d in s for r in (resolve_results[d],) if isinstance(r, models.Pkg)],)
            if m
        ]
        return [
            AddonMatch(folders=[{'name': f.name, 'version': f.version} for f in a], matches=m)
            for a, m in matches
        ] + [
            AddonMatch(folders=[{'name': f.name, 'version': f.version}], matches=[])
            for f in sorted(leftovers - frozenset(i for a, _ in matches for i in a))
        ]


class ReconcileInstalledCandidate(TypedDict):
    installed_addon: models.Pkg
    alternative_addons: list[models.Pkg]


@_register_method('get_reconcile_installed_candidates')
class GetReconcileInstalledCandidatesParams(_ProfileParamMixin, BaseParams):
    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> list[ReconcileInstalledCandidate]:
        manager = await managers.get_manager(self.profile)
        installed_pkgs = (
            models.Pkg.from_row_mapping(manager.database, p)
            for p in manager.database.execute(
                sa.select(db.pkg).order_by(sa.func.lower(db.pkg.c.name))
            )
            .mappings()
            .all()
        )
        defn_groups = await managers.run(
            self.profile, partial(Manager.find_equivalent_pkg_defns, pkgs=installed_pkgs)
        )
        resolve_results = await managers.run(
            self.profile,
            partial(Manager.resolve, defns=uniq(d for b in defn_groups.values() for d in b)),
        )
        return [
            {'installed_addon': p, 'alternative_addons': m}
            for p, s in defn_groups.items()
            for m in ([r for d in s for r in (resolve_results[d],) if isinstance(r, models.Pkg)],)
            if m
        ]


class DownloadProgressReport(TypedDict):
    defn: Defn
    progress: float


@_register_method('get_download_progress')
class GetDownloadProgressParams(_ProfileParamMixin, BaseParams):
    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> list[DownloadProgressReport]:
        return [
            {'defn': Defn.from_pkg(p), 'progress': r}
            for p, r in await managers.get_manager_download_progress(self.profile)
        ]


class GetVersionResult(TypedDict):
    installed_version: str
    new_version: str | None


@_register_method('meta/get_version')
class GetVersionParams(BaseParams):
    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> GetVersionResult:
        outdated, new_version = await is_outdated()
        return {
            'installed_version': __version__,
            'new_version': new_version if outdated else None,
        }


@_register_method('assist/open_url')
class OpenUrlParams(BaseParams):
    url: str

    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> None:
        click.launch(self.url)


@_register_method('assist/reveal_folder')
class RevealFolderParams(BaseParams):
    path_parts: typing.List[str]

    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> None:
        click.launch(os.path.join(*self.path_parts), locate=True)


class SelectFolderResult(TypedDict):
    selection: str | None


@_register_method('assist/select_folder')
class SelectFolderParams(BaseParams):
    initial_folder: typing.Optional[str]

    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> SelectFolderResult:
        if not app_window:
            raise RuntimeError('No app to bind to')
        try:
            (selection,) = app_window.select_folder_dialog('Select folder', self.initial_folder)
        except ValueError:
            selection = None
        return {'selection': selection}


class ConfirmDialogueResult(TypedDict):
    ok: bool


@_register_method('assist/confirm')
class ConfirmDialogueParams(BaseParams):
    title: str
    message: str

    async def respond(
        self, managers: _ManagerWorkQueue, app_window: toga.MainWindow | None
    ) -> ConfirmDialogueResult:
        if not app_window:
            raise RuntimeError('No app to bind to')
        ok = app_window.confirm_dialog(self.title, self.message)
        return {'ok': ok}


def _init_json_rpc_web_client(
    progress_reporters: set[tuple[Manager, models.Pkg, Callable[[], float]]],
) -> aiohttp.ClientSession:
    async def do_on_request_end(
        client_session: aiohttp.ClientSession,
        trace_config_ctx: Any,
        params: aiohttp.TraceRequestEndParams,
    ) -> None:
        trace_request_ctx: TraceRequestCtx = trace_config_ctx.trace_request_ctx
        if (
            trace_request_ctx
            and trace_request_ctx['report_progress'] == 'pkg_download'
            and params.response.content_length
            and aiohttp.hdrs.CONTENT_ENCODING not in params.response.headers
        ):
            content = params.response.content
            content_length = params.response.content_length
            entry = (
                trace_request_ctx['manager'],
                trace_request_ctx['pkg'],
                lambda: content.total_bytes / content_length,
            )
            progress_reporters.add(entry)
            content.on_eof(lambda: progress_reporters.remove(entry))

    trace_config = aiohttp.TraceConfig()
    trace_config.on_request_end.append(do_on_request_end)
    trace_config.freeze()
    return init_web_client(trace_configs=[trace_config])


class _ManagerWorkQueue:
    def __init__(self):
        self._loop = asyncio.get_running_loop()

        self._queue: asyncio.Queue[
            tuple[asyncio.Future[object], str, ManagerBoundCoroFn[..., object]]
        ] = asyncio.Queue()
        self._managers: dict[str, tuple[Manager, Callable[[], None]]] = {}
        self._progress_reporters: set[tuple[Manager, models.Pkg, Callable[[], float]]] = set()

        self._web_client = _init_json_rpc_web_client(self._progress_reporters)
        self.locks: LocksType = defaultdict(asyncio.Lock)

    async def cleanup(self):
        for _, close_db_conn in self._managers.values():
            close_db_conn()
        await self._web_client.close()

    def unload(self, profile: str):
        if profile in self._managers:
            _, close_db_conn = self._managers.pop(profile)
            close_db_conn()

    async def _schedule_item(
        self,
        future: asyncio.Future[object],
        profile: str,
        coro_fn: ManagerBoundCoroFn[_P, object],
    ):
        try:
            async with self.locks['modify profile', profile]:
                try:
                    manager, _ = self._managers[profile]
                except KeyError:
                    config = await _read_config(profile)
                    self._managers[profile] = Manager.from_config(config)
                    manager, _ = self._managers[profile]

            result = await coro_fn(manager)
        except BaseException as exc:
            future.set_exception(exc)
        else:
            future.set_result(result)

    async def listen(self):
        Manager.contextualise(web_client=self._web_client, locks=self.locks)
        while True:
            item = await self._queue.get()
            asyncio.create_task(self._schedule_item(*item))
            self._queue.task_done()

    async def run(self, profile: str, coro_fn: ManagerBoundCoroFn[..., _T]) -> _T:
        future = self._loop.create_future()
        self._queue.put_nowait((future, profile, coro_fn))
        return await asyncio.wait_for(future, None)

    async def get_manager(self, profile: str):
        async def get_manager(manager: Manager):
            return manager

        return await self.run(profile, get_manager)

    async def get_manager_download_progress(self, profile: str):
        manager = await self.get_manager(profile)
        return ((p, r()) for m, p, r in self._progress_reporters if m is manager)


async def create_app(app_window: toga.MainWindow | None = None):
    managers = _ManagerWorkQueue()

    async def start_managers(app: aiohttp.web.Application):
        managers_listen = asyncio.create_task(managers.listen())
        yield
        managers_listen.cancel()
        await managers.cleanup()

    async def get_index(request: aiohttp.web.Request):
        return aiohttp.web.Response(
            content_type='text/html',
            text=importlib.resources.read_text(frontend, 'index.html'),
        )

    async def get_static_file(request: aiohttp.web.Request):
        filename = request.path.lstrip('/')
        if filename.startswith('svelte-bundle.js'):
            content_type = 'application/javascript'
        elif filename == 'svelte-bundle.css':
            content_type = 'text/css'
        else:
            raise aiohttp.web.HTTPNotFound

        return aiohttp.web.Response(
            content_type=content_type,
            text=importlib.resources.read_text(frontend, filename),
        )

    def json_serialize(value: dict[str, Any]):
        return BaseModel.construct(**value).json()

    rpc_server = WsJsonRpcServer(
        json_serialize=json_serialize,
        middlewares=rpc_middlewares.DEFAULT_MIDDLEWARES,
    )
    rpc_server.add_methods([m.bind(n, managers, app_window) for n, m in methods])

    @aiohttp.web.middleware
    async def enforce_same_origin(
        request: aiohttp.web.Request,
        handler: Callable[[aiohttp.web.Request], Awaitable[aiohttp.web.StreamResponse]],
    ):
        if request.remote != LOCALHOST:
            raise aiohttp.web.HTTPUnauthorized

        if request.path.startswith('/api'):
            origin = request.headers.get(aiohttp.hdrs.ORIGIN)
            if origin is None:
                raise aiohttp.web.HTTPUnauthorized

            origin_url = URL(origin)
            if (
                origin_url.scheme != request.url.scheme
                or origin_url.host != request.url.host
                or origin_url.port != request.url.port
            ):
                raise aiohttp.web.HTTPUnauthorized

        return await handler(request)

    app = aiohttp.web.Application(middlewares=[enforce_same_origin])
    app.add_routes(
        [
            aiohttp.web.get('/', get_index),
            aiohttp.web.get(
                r'/svelte-bundle{extension:(?:\.css|\.js(?:\.map)?)}', get_static_file
            ),
            aiohttp.web.get('/api', rpc_server.handle_http_request),
        ]
    )
    app.cleanup_ctx.append(start_managers)
    app.freeze()
    return app


async def run_app(app: aiohttp.web.Application):
    "Fire up the server."
    app_runner = aiohttp.web.AppRunner(app)
    await app_runner.setup()
    assert app_runner.server  # Server is created during setup
    # By omitting the port ``loop.create_server`` will find an available port
    # to bind to - equivalent to creating a socket on port 0.
    server = await asyncio.get_running_loop().create_server(app_runner.server, LOCALHOST)
    host, port = server.sockets[0].getsockname()

    async def serve():
        try:
            # ``server_forever`` cleans up after the server when it's interrupted
            await server.serve_forever()
        finally:
            await app_runner.cleanup()

    return (
        URL.build(scheme='http', host=host, port=port),
        serve,
    )



"""
Replace any CF-specific strategies in the database.

Revision ID: 75f69831f74f
Revises: 9b8df9661cdb
Create Date: 2021-11-12 13:14:12.745467

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '75f69831f74f'
down_revision = '9b8df9661cdb'
branch_labels = None
depends_on = None


def upgrade():
    pkg_options = sa.table('pkg_options', sa.column('strategy', sa.String))
    op.execute(
        sa.update(pkg_options)
        .where(pkg_options.c.strategy.in_(['curse_latest_alpha', 'curse_latest_beta']))
        .values(strategy='latest')
    )


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



from __future__ import annotations

from collections.abc import Iterable, Iterator, Sequence
from functools import partial, reduce
from itertools import chain, product
import time
import typing
from typing import Any, ClassVar, TypeVar

from loguru import logger
from pydantic import BaseModel, Field, validator
from typing_extensions import Literal
from typing_extensions import NotRequired as N
from typing_extensions import Self, TypeAlias, TypedDict
from yarl import URL

from .common import Flavour
from .manager import Manager
from .utils import bucketise, chain_dict, gather
from .utils import run_in_thread as t
from .utils import shasum

_TAuras = TypeVar('_TAuras', bound='BaseAuras')
_ImportString = _Slug = str
_AuraGroup: TypeAlias = 'Sequence[tuple[Sequence[WeakAura], WagoApiResponse, _ImportString]]'

IMPORT_API_URL = URL('https://data.wago.io/api/raw/encoded')


class BaseAuras(
    BaseModel,
    arbitrary_types_allowed=True,
    json_encoders={URL: str},
):
    api_url: ClassVar[URL]
    filename: ClassVar[str]

    __root__: typing.Dict[_Slug, typing.List[Any]]

    @classmethod
    def from_lua_table(cls, lua_table: Any) -> Self:
        raise NotImplementedError


def _merge_auras(auras: Iterable[_TAuras]) -> dict[type[_TAuras], _TAuras]:
    "Merge auras of the same type."
    return {
        t: t(__root__=reduce(lambda a, b: {**a, **b}, (i.__root__ for i in a)))
        for t, a in bucketise(auras, key=type).items()
    }


class WeakAura(
    BaseModel,
    allow_population_by_field_name=True,
    arbitrary_types_allowed=True,
):
    id: str
    uid: str
    parent: typing.Optional[str]
    url: URL
    version: int

    @classmethod
    def from_lua_table(cls, lua_table: Any):
        url_string = lua_table.get('url')
        if url_string is not None:
            url = URL(url_string)
            if url.host == 'wago.io':
                return cls.parse_obj({**lua_table, 'url': url})

    @validator('url', pre=True)
    def _url_to_URL(cls, value: str | URL):
        if not isinstance(value, URL):
            value = URL(value)
        return value


class WeakAuras(BaseAuras):
    api_url = URL('https://data.wago.io/api/check/weakauras')
    filename = 'WeakAuras.lua'

    __root__: typing.Dict[_Slug, typing.List[WeakAura]]

    @classmethod
    def from_lua_table(cls, lua_table: Any):
        auras = (
            a for t in lua_table['displays'].values() for a in (WeakAura.from_lua_table(t),) if a
        )
        sorted_auras = sorted(auras, key=lambda a: a.id)
        return cls(__root__=bucketise(sorted_auras, key=lambda a: a.url.parts[1]))


class Plateroo(WeakAura):
    id: str = Field(alias='Name')
    uid = ''


class Plateroos(BaseAuras):
    api_url = URL('https://data.wago.io/api/check/plater')
    filename = 'Plater.lua'

    __root__: typing.Dict[_Slug, typing.List[Plateroo]]

    @classmethod
    def from_lua_table(cls, lua_table: Any):
        auras = (
            a
            for n, p in lua_table['profiles'].items()
            for t in chain(
                ({**p, 'Name': f'__profile_{n}__'},),
                p.get('script_data') or (),
                p.get('hook_data') or (),
            )
            for a in (Plateroo.from_lua_table(t),)
            if a
        )
        sorted_auras = sorted(auras, key=lambda a: a.id)
        return cls(__root__={a.url.parts[1]: [a] for a in sorted_auras})


class WagoApiResponse(TypedDict):
    _id: str  # +   # Alphanumeric ID
    name: str  # +  # User-facing name
    slug: str  # +  # Slug if it has one; otherwise same as ``_id``
    url: str
    created: str  # ISO datetime
    modified: str  # ISO datetime
    game: str  # "classic" or xpac, e.g. "bfa"
    username: N[str]  # +  # Author username
    version: int  # +   # Version counter, incremented with every update
    # Semver auto-generated from ``version`` - for presentation only
    versionString: str
    changelog: WagoApiResponse_Changelog  # +
    forkOf: N[str]  # Only present on forks
    regionType: N[str]  # Only present on WAs


class WagoApiResponse_Changelog(TypedDict):
    format: N[Literal['bbcode', 'markdown']]
    text: N[str]


class WaCompanionBuilder:
    """A WeakAuras Companion port for shellfolk."""

    def __init__(self, manager: Manager) -> None:
        self.manager = manager

        self.access_token = self.manager.config.global_config.access_tokens.wago

        output_folder = self.manager.config.plugin_dir / __name__
        self.addon_zip_path = output_folder / 'WeakAurasCompanion.zip'
        self.changelog_path = output_folder / 'CHANGELOG.md'
        self.checksum_txt_path = output_folder / 'checksum.txt'

    @staticmethod
    def extract_auras(model: type[_TAuras], source: str) -> _TAuras:
        from ._custom_slpp import loads

        lua_table = loads(f'{{{source}}}')
        (aura_table,) = lua_table.values()
        return model.from_lua_table(aura_table)

    def extract_installed_auras(self) -> Iterator[WeakAuras | Plateroos]:
        flavour_root = self.manager.config.addon_dir.parents[1]
        saved_vars_by_account = flavour_root.glob('WTF/Account/*/SavedVariables')
        for saved_vars, model in product(
            saved_vars_by_account,
            [WeakAuras, Plateroos],
        ):
            file = saved_vars / model.filename
            if not file.exists():
                logger.info(f'{file} not found')
            else:
                content = file.read_text(encoding='utf-8-sig', errors='replace')
                aura_group_cache = self.manager.config.global_config.cache_dir / shasum(content)
                if aura_group_cache.exists():
                    logger.info(f'loading {file} from cache at {aura_group_cache}')
                    aura_groups = model.parse_file(aura_group_cache)
                else:
                    start = time.perf_counter()
                    aura_groups = self.extract_auras(model, content)
                    logger.debug(
                        f'extracted {model.__name__} in {time.perf_counter() - start:.3f}s'
                    )
                    aura_group_cache.write_text(aura_groups.json(), encoding='utf-8')
                yield aura_groups

    async def _fetch_wago_metadata(self, api_url: URL, aura_ids: Iterable[str]):
        async with self.manager.web_client.get(
            api_url.with_query(ids=','.join(aura_ids)),
            {'minutes': 30},
            label='Fetching aura metadata',
            headers={
                'api-key': self.access_token.get_secret_value()
                if self.access_token is not None
                else ''
            },
        ) as response:
            metadata: list[WagoApiResponse]
            if response.status == 404:
                metadata = []
                return metadata
            response.raise_for_status()
            metadata = await response.json()
            return sorted(metadata, key=lambda r: r['slug'])

    async def _fetch_wago_import_string(self, aura: WagoApiResponse):
        async with self.manager.web_client.get(
            IMPORT_API_URL.with_query(id=aura['_id']).with_fragment(str(aura['version'])),
            {'days': 30},
            headers={
                'api-key': self.access_token.get_secret_value()
                if self.access_token is not None
                else ''
            },
            label=f"Fetching aura '{aura['slug']}'",
        ) as response:
            response.raise_for_status()
            return await response.text()

    async def get_remote_auras(self, auras: BaseAuras) -> _AuraGroup:
        if not auras.__root__:
            return []

        metadata = await self._fetch_wago_metadata(auras.api_url, auras.__root__)
        import_strings = await gather(self._fetch_wago_import_string(r) for r in metadata)
        return [(auras.__root__[r['slug']], r, i) for r, i in zip(metadata, import_strings)]

    def _checksum(self) -> str:
        from hashlib import sha256

        return sha256(self.addon_zip_path.read_bytes()).hexdigest()

    async def get_checksum(self) -> str:
        return await t(self.checksum_txt_path.read_text)(encoding='utf-8')

    def _get_toc_number(self) -> str:
        game_flavour: Flavour = self.manager.config.game_flavour
        if game_flavour is Flavour.retail:
            return '90105'
        elif game_flavour is Flavour.vanilla_classic:
            return '11400'
        elif game_flavour is Flavour.burning_crusade_classic:
            return '20502'

    def _generate_addon(
        self, auras: Iterable[tuple[type[WeakAuras | Plateroos], _AuraGroup]]
    ) -> None:
        from importlib.resources import read_text
        from zipfile import ZipFile, ZipInfo

        from jinja2 import Environment, FunctionLoader

        from . import wa_templates

        jinja_env = Environment(
            trim_blocks=True,
            lstrip_blocks=True,
            loader=FunctionLoader(partial(read_text, wa_templates)),
        )
        aura_dict = chain_dict((WeakAuras, Plateroos), (), auras)

        self.addon_zip_path.parent.mkdir(exist_ok=True)
        with ZipFile(self.addon_zip_path, 'w') as file:

            def write_tpl(filename: str, ctx: dict[str, Any]):
                # Not using a plain string as the first argument to ``writestr``
                # 'cause the timestamp would be set to the current time
                # which would render the build unreproducible
                zip_info = ZipInfo(filename=f'WeakAurasCompanion/{filename}')
                file.writestr(zip_info, jinja_env.get_template(filename).render(ctx))

            write_tpl(
                'data.lua',
                {
                    'weakauras': [
                        (
                            metadata['slug'],
                            {
                                'name': metadata['name'],
                                'author': metadata.get('username', '__unknown__'),
                                'encoded': import_string,
                                'wagoVersion': metadata['version'],
                                # ``wagoSemver`` is supposed to be the ``versionString``
                                # from Wago but there is a bug where the ``version``
                                # is sometimes not appended to the semver.
                                # The Companion add-on's version is derived from its checksum
                                # so if ``wagoSemver`` were to change between requests
                                # we'd be triggering spurious updates in instawow.
                                'wagoSemver': metadata['version'],
                                'versionNote': metadata['changelog'].get('text', ''),
                            },
                        )
                        for _, metadata, import_string in aura_dict[WeakAuras]
                    ],
                    # Maps internal UIDs of top-level auras to IDs or slugs on Wago
                    'weakaura_uids': [
                        (a.uid, a.url.parts[1])
                        for existing_auras, _, _ in aura_dict[WeakAuras]
                        for a in (
                            next((i for i in existing_auras if not i.parent), existing_auras[0]),
                        )
                    ],
                    # Maps local names to IDs or slugs on Wago
                    'weakaura_ids': [
                        (a.id, a.url.parts[1])
                        for existing_auras, _, _ in aura_dict[WeakAuras]
                        for a in existing_auras
                    ],
                    'plateroos': [
                        (
                            metadata['slug'],
                            {
                                'name': metadata['name'],
                                'author': metadata.get('username', '__unknown__'),
                                'encoded': import_string,
                                'wagoVersion': metadata['version'],
                                'wagoSemver': metadata['version'],
                                'versionNote': metadata['changelog'].get('text', ''),
                            },
                        )
                        for _, metadata, import_string in aura_dict[Plateroos]
                    ],
                    'plater_ids': [
                        (a.id, a.url.parts[1])
                        for existing_auras, _, _ in aura_dict[Plateroos]
                        for a in existing_auras
                    ],
                },
            )
            write_tpl('init.lua', {})
            write_tpl(
                'WeakAurasCompanion.toc',
                {'interface': self._get_toc_number()},
            )

        self.changelog_path.write_text(
            jinja_env.get_template(self.changelog_path.name).render(
                {
                    'changelog_entries': [
                        (
                            a.id,
                            a.url.parent,
                            metadata['version'],
                            metadata['changelog'].get('text') or 'n/a',
                        )
                        for v in aura_dict.values()
                        for existing_auras, metadata, _ in v
                        for a in (
                            next((i for i in existing_auras if not i.parent), existing_auras[0]),
                        )
                        if a.version != metadata['version']
                    ]
                }
            )
            or 'n/a',
            encoding='utf-8',
        )

        self.checksum_txt_path.write_text(
            self._checksum(),
            encoding='utf-8',
        )

    async def build(self) -> None:
        installed_auras = await t(list)(self.extract_installed_auras())
        installed_auras_by_type = _merge_auras(installed_auras)
        aura_groups = await gather(map(self.get_remote_auras, installed_auras_by_type.values()))
        await t(self._generate_addon)(zip(installed_auras_by_type, aura_groups))



from __future__ import annotations

from collections.abc import Callable
from functools import lru_cache
from io import BytesIO
import json
import os
from pathlib import Path
import re
from typing import Any
from zipfile import ZipFile

import aiohttp
from aresponses import ResponsesMockServer
import pytest

from instawow import __version__
from instawow.common import Flavour
from instawow.config import Config, GlobalConfig
from instawow.manager import Manager, init_web_client

inf = float('inf')

FIXTURES = Path(__file__).parent / 'fixtures'


def pytest_addoption(parser: pytest.Parser):
    parser.addoption('--iw-no-mock-http', action='store_true')


def should_mock(fn: Callable[..., object]):
    import inspect
    import warnings

    def wrapper(request: pytest.FixtureRequest):
        if request.config.getoption('--iw-no-mock-http'):
            warnings.warn('not mocking')
            return None
        elif any(m.name == 'iw_no_mock_http' for m in request.node.iter_markers()):
            return None

        args = (request.getfixturevalue(p) for p in inspect.signature(fn).parameters)
        return fn(*args)

    return wrapper


@lru_cache(maxsize=None)
def load_fixture(filename: str):
    return (FIXTURES / filename).read_bytes()


@lru_cache(maxsize=None)
def load_json_fixture(filename: str):
    return json.loads(load_fixture(filename))


@lru_cache(maxsize=None)
def make_addon_zip(*folders: str):
    buffer = BytesIO()
    with ZipFile(buffer, 'w') as file:
        for folder in folders:
            file.writestr(f'{folder}/{folder}.toc', b'')

    return buffer.getvalue()


@pytest.fixture(scope='session', autouse=True)
def iw_temp_dir(tmp_path_factory: pytest.TempPathFactory):
    temp_dir = tmp_path_factory.mktemp('temp') / 'instawow'
    os.environ['INSTAWOW_TEMP_DIR'] = str(temp_dir)
    return temp_dir


@pytest.fixture
def iw_global_config_values(tmp_path: Path, iw_temp_dir: Path):
    return {'temp_dir': iw_temp_dir, 'config_dir': tmp_path / 'config'}


@pytest.fixture(params=Flavour)
def iw_config_values(request: Any, tmp_path: Path):
    addons = tmp_path / 'wow' / 'interface' / 'addons'
    addons.mkdir(parents=True)
    return {'profile': '__default__', 'addon_dir': addons, 'game_flavour': request.param}


@pytest.fixture
def iw_config(iw_config_values: dict[str, Any], iw_global_config_values: dict[str, Any]):
    global_config = GlobalConfig(**iw_global_config_values).write()
    return Config(global_config=global_config, **iw_config_values).write()


@pytest.fixture
async def iw_web_client():
    async with init_web_client() as web_client:
        yield web_client


@pytest.fixture
def iw_manager(iw_config: Config, iw_web_client: aiohttp.ClientSession):
    Manager.contextualise(web_client=iw_web_client)
    manager, close_db_conn = Manager.from_config(iw_config)
    yield manager
    close_db_conn()


@pytest.fixture(autouse=True)
@should_mock
def mock_aiohttp_requests(aresponses: ResponsesMockServer):
    aresponses.add(
        'pypi.org',
        '/pypi/instawow/json',
        'get',
        {'info': {'version': __version__}},
        repeat=inf,
    )

    aresponses.add(
        'raw.githubusercontent.com',
        '/layday/instawow-data/data/base-catalogue-v5.compact.json',
        'get',
        load_json_fixture('base-catalogue-v5.compact.json'),
        repeat=inf,
    )

    aresponses.add(
        'addons-ecs.forgesvc.net',
        '/api/v2/addon',
        'post',
        load_json_fixture('curse-addon--all.json'),
        repeat=inf,
    )
    aresponses.add(
        'addons-ecs.forgesvc.net',
        '/api/v2/addon/20338/files',
        'get',
        load_json_fixture('curse-addon-files.json'),
        repeat=inf,
    )
    aresponses.add(
        'addons-ecs.forgesvc.net',
        re.compile(r'^/api/v2/addon/20338/file/(\d+)/changelog'),
        'get',
        aresponses.Response(text=load_fixture('curse-addon-changelog.txt').decode()),
        repeat=inf,
    )
    aresponses.add(
        'edge.forgecdn.net',
        aresponses.ANY,
        'get',
        aresponses.Response(body=make_addon_zip('Molinari')),
        repeat=inf,
    )

    aresponses.add(
        'api.mmoui.com',
        '/v3/game/WOW/filelist.json',
        'get',
        load_json_fixture('wowi-filelist.json'),
        repeat=inf,
    )
    aresponses.add(
        'api.mmoui.com',
        re.compile(r'^/v3/game/WOW/filedetails/'),
        'get',
        load_json_fixture('wowi-filedetails.json'),
        repeat=inf,
    )
    aresponses.add(
        'cdn.wowinterface.com',
        aresponses.ANY,
        'get',
        aresponses.Response(body=make_addon_zip('Molinari')),
        repeat=inf,
    )

    aresponses.add(
        'www.tukui.org',
        '/api.php?ui=tukui',
        'get',
        load_json_fixture('tukui-ui--tukui.json'),
        match_querystring=True,
        repeat=inf,
    )
    aresponses.add(
        'www.tukui.org',
        '/api.php?ui=elvui',
        'get',
        load_json_fixture('tukui-ui--elvui.json'),
        match_querystring=True,
        repeat=inf,
    )
    aresponses.add(
        'www.tukui.org',
        '/api.php?addons=all',
        'get',
        load_json_fixture('tukui-retail-addons.json'),
        match_querystring=True,
        repeat=inf,
    )
    aresponses.add(
        'www.tukui.org',
        '/api.php?classic-addons=all',
        'get',
        load_json_fixture('tukui-classic-addons.json'),
        match_querystring=True,
        repeat=inf,
    )
    aresponses.add(
        'www.tukui.org',
        '/api.php?classic-tbc-addons=all',
        'get',
        load_json_fixture('tukui-classic-tbc-addons.json'),
        match_querystring=True,
        repeat=inf,
    )
    aresponses.add(
        'www.tukui.org',
        '/api.php',
        'get',
        '',
        repeat=inf,
    )
    aresponses.add(
        'www.tukui.org',
        re.compile(r'^/downloads/tukui'),
        'get',
        aresponses.Response(body=make_addon_zip('Tukui')),
        repeat=inf,
    )
    aresponses.add(
        'www.tukui.org',
        '/addons.php?download=1',
        'get',
        aresponses.Response(body=make_addon_zip('ElvUI_MerathilisUI')),
        match_querystring=True,
        repeat=inf,
    )
    aresponses.add(
        'www.tukui.org',
        re.compile(r'/classic-(?:tbc-)?addons\.php\?download=1'),
        'get',
        aresponses.Response(body=make_addon_zip('Tukui')),
        match_querystring=True,
        repeat=inf,
    )

    aresponses.add(
        'api.github.com',
        '/repos/nebularg/PackagerTest',
        'get',
        load_json_fixture('github-repo-release-json.json'),
        repeat=inf,
    )
    aresponses.add(
        'api.github.com',
        '/repos/nebularg/PackagerTest/releases/latest',
        'get',
        load_json_fixture('github-release-release-json.json'),
        repeat=inf,
    )
    aresponses.add(
        'github.com',
        '/nebularg/PackagerTest/releases/download/v1.9.7/release.json',
        'get',
        load_json_fixture('github-release-release-json-release-json.json'),
        repeat=inf,
    )
    aresponses.add(
        'api.github.com',
        re.compile(r'^/repos/p3lim-wow/Molinari$', re.IGNORECASE),
        'get',
        load_json_fixture('github-repo-molinari.json'),
        repeat=inf,
    )
    aresponses.add(
        'api.github.com',
        re.compile(r'^/repos/p3lim-wow/Molinari/releases/latest$', re.IGNORECASE),
        'get',
        load_json_fixture('github-release-molinari.json'),
        repeat=inf,
    )
    aresponses.add(
        'github.com',
        re.compile(
            r'^/p3lim-wow/Molinari/releases/download/90105\.81-Release/release\.json$',
            re.IGNORECASE,
        ),
        'get',
        load_json_fixture('github-release-molinari-release-json.json'),
        repeat=inf,
    )
    aresponses.add(
        'api.github.com',
        '/repos/AdiAddons/AdiBags',
        'get',
        load_json_fixture('github-repo-no-releases.json'),
        repeat=inf,
    )
    aresponses.add(
        'api.github.com',
        '/repos/AdiAddons/AdiBags/releases/latest',
        'get',
        aresponses.Response(body=b'', status=404),
        repeat=inf,
    )
    aresponses.add(
        'api.github.com',
        '/repos/AdiAddons/AdiButtonAuras/releases/tags/2.0.19',
        'get',
        load_json_fixture('github-release-no-assets.json'),
        repeat=inf,
    )
    aresponses.add(
        'api.github.com',
        '/repos/layday/foobar',
        'get',
        aresponses.Response(body=b'', status=404),
        repeat=inf,
    )
    aresponses.add(
        'github.com',
        re.compile(r'^(/[^/]*){2}/releases/download'),
        'get',
        aresponses.Response(body=make_addon_zip('Foo')),
        repeat=inf,
    )

    aresponses.add(
        'github.com',
        '/login/device/code',
        'post',
        load_json_fixture('github-oauth-login-device-code.json'),
        repeat=inf,
    )
    aresponses.add(
        'github.com',
        '/login/oauth/access_token',
        'post',
        load_json_fixture('github-oauth-login-access-token.json'),
        repeat=inf,
    )



"""
Replace the ``pkg_dep`` UNIQUE constraint with a pk.

Revision ID: e4921edb1154
Revises: 43aa3610e92a
Create Date: 2020-09-08 14:39:34.957011

"""
from contextlib import suppress

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'e4921edb1154'
down_revision = '43aa3610e92a'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('pkg_dep', schema=None) as batch_op:
        batch_op.drop_column('_id')
        batch_op.create_primary_key('pk_pkg_dep', ['id', 'pkg_source', 'pkg_id'])

    # ``UNIQUE`` constraint might've previously not been applied correctly
    with suppress(ValueError), op.batch_alter_table('pkg_dep', schema=None) as batch_op:
        batch_op.drop_constraint('uq_id_per_foreign_key_constr', type_='unique')


def downgrade():
    with op.batch_alter_table('pkg_dep', schema=None) as batch_op:
        batch_op.add_column(sa.Column('_id', sa.INTEGER(), nullable=False))
        batch_op.drop_constraint('pk_pkg_dep', type_='primary')
        batch_op.create_primary_key('pk_pkg_dep', ['_id'])
        batch_op.create_unique_constraint(
            'uq_id_per_foreign_key_constr', ['id', 'pkg_origin', 'pkg_id']
        )



from __future__ import annotations

from contextlib import contextmanager
from datetime import datetime
import typing
from typing import Any, Mapping

from pydantic import BaseModel
import sqlalchemy as sa
import sqlalchemy.future as sa_future

from . import db
from .common import Strategy


@contextmanager
def _faux_transaction(connection: sa_future.Connection):
    try:
        yield
    except BaseException:
        connection.rollback()
        raise
    else:
        connection.commit()


class _PkgOptions(BaseModel):
    strategy: Strategy


class _PkgFolder(BaseModel):
    name: str


class _PkgDep(BaseModel):
    id: str


class _PkgLoggedVersion(BaseModel):
    version: str
    install_time: datetime


class Pkg(BaseModel):
    source: str
    id: str
    slug: str
    name: str
    description: str
    url: str
    download_url: str
    date_published: datetime
    version: str
    changelog_url: str
    options: _PkgOptions  # pkg_options
    folders: typing.List[_PkgFolder] = []  # pkg_folder
    deps: typing.List[_PkgDep] = []  # pkg_dep
    logged_versions: typing.List[_PkgLoggedVersion] = []  # pkg_version_log

    @classmethod
    def from_row_mapping(
        cls, connection: sa_future.Connection, row_mapping: Mapping[str, Any]
    ) -> Pkg:
        source_and_id = {'pkg_source': row_mapping['source'], 'pkg_id': row_mapping['id']}
        return cls.parse_obj(
            {
                **row_mapping,
                'options': connection.execute(
                    sa.select(db.pkg_options.c.strategy).filter_by(**source_and_id)
                )
                .mappings()
                .one(),
                'folders': connection.execute(
                    sa.select(db.pkg_folder.c.name).filter_by(**source_and_id)
                )
                .mappings()
                .all(),
                'deps': connection.execute(sa.select(db.pkg_dep.c.id).filter_by(**source_and_id))
                .mappings()
                .all(),
                'logged_versions': connection.execute(
                    sa.select(db.pkg_version_log)
                    .filter_by(**source_and_id)
                    .order_by(db.pkg_version_log.c.install_time.desc())
                    .limit(10)
                )
                .mappings()
                .all(),
            }
        )

    def insert(self, connection: sa_future.Connection) -> None:
        values = self.dict()
        source_and_id = {'pkg_source': values['source'], 'pkg_id': values['id']}
        with _faux_transaction(connection):
            connection.execute(sa.insert(db.pkg), [values])
            connection.execute(
                sa.insert(db.pkg_folder), [{**f, **source_and_id} for f in values['folders']]
            )
            connection.execute(sa.insert(db.pkg_options), [{**values['options'], **source_and_id}])
            if values['deps']:
                connection.execute(
                    sa.insert(db.pkg_dep), [{**d, **source_and_id} for d in values['deps']]
                )
            connection.execute(
                sa.insert(db.pkg_version_log).prefix_with('OR IGNORE'),
                [{'version': values['version'], **source_and_id}],
            )

    def delete(self, connection: sa_future.Connection) -> None:
        source_and_id = {'pkg_source': self.source, 'pkg_id': self.id}
        with _faux_transaction(connection):
            connection.execute(sa.delete(db.pkg_dep).filter_by(**source_and_id))
            connection.execute(sa.delete(db.pkg_folder).filter_by(**source_and_id))
            connection.execute(sa.delete(db.pkg_options).filter_by(**source_and_id))
            connection.execute(sa.delete(db.pkg).filter_by(source=self.source, id=self.id))

    # Make the model hashable again
    __eq__ = object.__eq__
    __hash__ = object.__hash__  # type: ignore


class PkgList(BaseModel):
    __root__: typing.List[Pkg]



from __future__ import annotations

from pathlib import Path
import sys
from typing import Any

import pytest

from instawow.common import Flavour
from instawow.config import Config, GlobalConfig


def test_env_vars_have_prio(
    monkeypatch: pytest.MonkeyPatch,
    iw_global_config_values: dict[str, Any],
    iw_config_values: dict[str, Any],
):
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', '/foo')
    monkeypatch.setenv('INSTAWOW_GAME_FLAVOUR', 'classic')

    global_config = GlobalConfig(**iw_global_config_values)
    config = Config(global_config=global_config, **iw_config_values)

    assert global_config.config_dir == Path('/foo').resolve()
    assert config.game_flavour is Flavour.burning_crusade_classic


def test_config_dir_is_populated(
    iw_global_config_values: dict[str, Any],
    iw_config_values: dict[str, Any],
):
    global_config = GlobalConfig(**iw_global_config_values)
    config = Config(global_config=global_config, **iw_config_values).write()
    assert {i.name for i in config.profile_dir.iterdir()} == {'config.json', 'logs', 'plugins'}


def test_read_profile_from_nonexistent_config_dir_raises(
    iw_global_config_values: dict[str, Any],
):
    global_config = GlobalConfig(config_dir=iw_global_config_values['config_dir'])
    with pytest.raises(FileNotFoundError):
        Config.read(global_config, '__default__')


def test_init_with_nonexistent_addon_dir_raises(
    iw_global_config_values: dict[str, Any],
    iw_config_values: dict[str, Any],
):
    global_config = GlobalConfig(**iw_global_config_values).write()
    with pytest.raises(ValueError, match='not a writable directory'):
        Config(global_config=global_config, **{**iw_config_values, 'addon_dir': 'foo'})


@pytest.mark.skipif(
    sys.platform == 'win32',
    reason='requires absence of platform-specific env var',
)
def test_default_config_dir_is_platform_appropriate(
    monkeypatch: pytest.MonkeyPatch,
):
    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'linux')
        config_dir = GlobalConfig().config_dir
        assert config_dir == Path.home() / '.config/instawow'

        patcher.setenv('XDG_CONFIG_HOME', '/foo')
        config_dir = GlobalConfig().config_dir
        assert config_dir == Path('/foo/instawow')

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'darwin')
        config_dir = GlobalConfig().config_dir
        assert config_dir == Path.home() / 'Library/Application Support/instawow'


@pytest.mark.skipif(
    sys.platform != 'win32',
    reason='requires presence of platform-specific env var',
)
def test_default_config_dir_is_unplatform_appropriate(
    monkeypatch: pytest.MonkeyPatch,
):
    assert GlobalConfig().config_dir == Path.home() / 'AppData/Roaming/instawow'
    monkeypatch.delenv('APPDATA')
    assert GlobalConfig().config_dir == Path.home() / 'instawow'


def test_can_infer_flavour_from_path():
    infer = Config.infer_flavour
    assert infer('wowzerz/_classic_/Interface/AddOns') is Flavour.burning_crusade_classic
    assert infer('/foo/bar/_classic_beta_/Interface/AddOns') is Flavour.burning_crusade_classic
    assert infer('/foo/bar/_classic_ptr_/Interface/AddOns') is Flavour.burning_crusade_classic
    assert infer('_classic_era_/Interface/AddOns') is Flavour.vanilla_classic
    assert infer('_classic_era_ptr_/Interface/AddOns') is Flavour.vanilla_classic
    assert infer('wowzerz/_retail_/Interface/AddOns') is Flavour.retail
    assert infer('anything goes') is Flavour.retail


def test_can_list_profiles(
    monkeypatch: pytest.MonkeyPatch,
    iw_global_config_values: dict[str, Any],
    iw_config_values: dict[str, Any],
):
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', str(iw_global_config_values['config_dir']))

    global_config = GlobalConfig()
    assert global_config.list_profiles() == []

    Config(global_config=global_config, **iw_config_values).write()
    Config(global_config=global_config, **{**iw_config_values, 'profile': 'foo'}).write()
    assert set(global_config.list_profiles()) == {'__default__', 'foo'}


def test_can_delete_profile(
    iw_global_config_values: dict[str, Any],
    iw_config_values: dict[str, Any],
):
    global_config = GlobalConfig(**iw_global_config_values).write()
    config = Config(global_config=global_config, **iw_config_values).write()
    assert config.profile_dir.exists()
    config.delete()
    assert not config.profile_dir.exists()


def test_can_omit_global_config(
    monkeypatch: pytest.MonkeyPatch,
    iw_global_config_values: dict[str, Any],
    iw_config_values: dict[str, Any],
):
    global_config = GlobalConfig(**iw_global_config_values)
    Config(global_config=global_config, **iw_config_values).write()

    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', str(iw_global_config_values['config_dir']))
    config = Config.read(None, iw_config_values['profile'])
    assert config.global_config.dict() == global_config.dict()



# Adapted from https://github.com/r0x0r/pywebview/blob/master/webview/platforms/edgechromium.py.

from __future__ import annotations

import importlib.resources
import os

import clr
from toga_winforms.libs import Uri
from toga_winforms.widgets.base import Widget

from . import webview2

core_lib = importlib.resources.path(webview2, 'Microsoft.Web.WebView2.Core.dll').__enter__()
winforms_lib = importlib.resources.path(
    webview2, 'Microsoft.Web.WebView2.WinForms.dll'
).__enter__()

clr.AddReference(os.fspath(core_lib))
clr.AddReference(os.fspath(winforms_lib))

from Microsoft.Web.WebView2.Core import CoreWebView2Environment, WebView2RuntimeNotFoundException
from Microsoft.Web.WebView2.WinForms import WebView2


def is_webview2_installed():
    try:
        CoreWebView2Environment.GetAvailableBrowserVersionString()
    except WebView2RuntimeNotFoundException:
        return False
    else:
        return True


class WebView2Widget(Widget):
    def create(self):
        self.native = WebView2()
        self.native.interface = self.interface

        def on_initialization_completed(sender, event_args):
            pass

        self.native.CoreWebView2InitializationCompleted += on_initialization_completed

    def set_on_key_down(self, handler: object) -> None:
        pass

    def set_on_webview_load(self, handler: object) -> None:
        pass

    def set_url(self, value: str | None) -> None:
        if value is not None:
            self.native.Source = Uri(value)

    def set_content(self, root_url: str, content: str) -> None:
        raise NotImplementedError

    def set_user_agent(self, value: str) -> None:
        pass

    async def evaluate_javascript(self, javascript: str) -> None:
        raise NotImplementedError

    def invoke_javascript(self, javascript: str) -> None:
        self.native.ExecuteScriptAsync(javascript)


class Factory:
    WebView = WebView2Widget



# pyright: reportUnusedFunction=false

from __future__ import annotations

import asyncio
from collections.abc import Sequence
from functools import partial
from typing import Any

from prompt_toolkit.application import Application
from prompt_toolkit.document import Document
from prompt_toolkit.formatted_text.html import HTML
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent
from prompt_toolkit.keys import Keys
from prompt_toolkit.shortcuts.progress_bar import ProgressBar, formatters
from prompt_toolkit.styles import Style
from prompt_toolkit.validation import ValidationError, Validator
import pydantic
from questionary import Choice
from questionary import confirm as _confirm
from questionary import path as _path
from questionary.prompts.common import InquirerControl, Separator, create_inquirer_layout
from questionary.question import Question

from . import models


class PydanticValidator(Validator):
    "One-off validators for Pydantic model fields."

    def __init__(self, model: type[pydantic.BaseModel], field: str) -> None:
        self.model = model
        self.field = field

    def validate(self, document: Document) -> None:
        try:
            self.model.parse_obj({self.field: document.text})
        except pydantic.ValidationError as error:
            error_at_loc = next((e for e in error.errors() if e['loc'] == (self.field,)), None)
            if error_at_loc:
                raise ValidationError(0, error_at_loc['msg'])


class PkgChoice(Choice):
    def __init__(self, *args: Any, pkg: models.Pkg, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.pkg = pkg


qstyle = Style(
    [
        ('qmark', 'fg:ansicyan'),
        ('question', 'bold'),
        ('answer', 'fg: nobold'),
        ('skipped-answer', 'fg:ansiyellow'),
        ('highlight-sub', 'fg:ansimagenta'),
    ]
)


SKIP = ()
skip = Choice([('', 'skip')], SKIP)

confirm = partial(_confirm, qmark='?', style=qstyle)
path = partial(_path, qmark='>', style=qstyle)


def checkbox(message: str, choices: Sequence[Choice], **inquirer_kwargs: Any) -> Question:
    def get_prompt_tokens():
        tokens: list[tuple[str, str]] = [('class:question', message)]
        if ic.is_answered:
            tokens.append(('class:answer', '  done'))
        else:
            tokens.append(
                (
                    'class:instruction',
                    '  (use arrow keys to move, <space> to select, <o> to view in your browser)',
                )
            )
        return tokens

    ic = InquirerControl(
        choices,
    )
    bindings = KeyBindings()

    @bindings.add(Keys.ControlQ, eager=True)
    @bindings.add(Keys.ControlC, eager=True)
    def abort(event: KeyPressEvent):
        event.app.exit(exception=KeyboardInterrupt, style='class:aborting')

    @bindings.add(' ', eager=True)
    def toggle(event: KeyPressEvent):
        pointed_choice = ic.get_pointed_at().value
        if pointed_choice in ic.selected_options:
            ic.selected_options.remove(pointed_choice)
        else:
            ic.selected_options.append(pointed_choice)

    @bindings.add('i', eager=True)
    def invert(event: KeyPressEvent):
        inverted_selection = [
            c.value
            for c in ic.choices
            if not isinstance(c, Separator)
            and c.value not in ic.selected_options
            and not c.disabled
        ]
        ic.selected_options = inverted_selection

    @bindings.add(Keys.Down, eager=True)
    @bindings.add('j', eager=True)
    def move_cursor_down(event: KeyPressEvent):
        ic.select_next()
        while not ic.is_selection_valid():
            ic.select_next()

    @bindings.add(Keys.Up, eager=True)
    @bindings.add('k', eager=True)
    def move_cursor_up(event: KeyPressEvent):
        ic.select_previous()
        while not ic.is_selection_valid():
            ic.select_previous()

    @bindings.add(Keys.ControlM, eager=True)
    def set_answer(event: KeyPressEvent):
        ic.is_answered = True
        event.app.exit(result=[c.value for c in ic.get_selected_values()])

    @bindings.add('o', eager=True)
    def open_url(event: KeyPressEvent):
        choice = ic.get_pointed_at()
        if isinstance(choice, PkgChoice):
            import webbrowser

            webbrowser.open(choice.pkg.url)

    @bindings.add(Keys.Any)
    def default(event: KeyPressEvent):
        # Disallow inserting other text
        pass

    layout = create_inquirer_layout(ic, get_prompt_tokens, **inquirer_kwargs)
    return Question(Application(layout=layout, key_bindings=bindings, style=qstyle))


def select(
    message: str,
    choices: Sequence[str] | Sequence[Choice],
    initial_choice: str | Choice | None = None,
    **inquirer_kwargs: Any,
) -> Question:
    def get_prompt_tokens():
        tokens: list[tuple[str, str]] = [('class:qmark', '- '), ('class:question', message)]
        if ic.is_answered:
            answer = ic.get_pointed_at()
            title = answer.title
            assert title
            tokens.extend(
                [
                    ('', ' '),
                    (
                        'class:skipped-answer' if answer is skip else 'class:answer',
                        ''.join(t[1] for t in title) if isinstance(title, list) else title,
                    ),
                ]
            )
        return tokens

    ic = InquirerControl(
        choices,
        use_indicator=False,
        initial_choice=initial_choice,
    )
    bindings = KeyBindings()

    @bindings.add(Keys.ControlQ, eager=True)
    @bindings.add(Keys.ControlC, eager=True)
    def abort(event: KeyPressEvent):
        event.app.exit(exception=KeyboardInterrupt, style='class:aborting')

    @bindings.add(Keys.Down, eager=True)
    @bindings.add('j', eager=True)
    def move_cursor_down(event: KeyPressEvent):
        ic.select_next()
        while not ic.is_selection_valid():
            ic.select_next()

    @bindings.add(Keys.Up, eager=True)
    @bindings.add('k', eager=True)
    def move_cursor_up(event: KeyPressEvent):
        ic.select_previous()
        while not ic.is_selection_valid():
            ic.select_previous()

    @bindings.add(Keys.ControlM, eager=True)
    def set_answer(event: KeyPressEvent):
        ic.is_answered = True
        event.app.exit(result=ic.get_pointed_at().value)

    @bindings.add('o', eager=True)
    def open_url(event: KeyPressEvent):
        choice = ic.get_pointed_at()
        if isinstance(choice, PkgChoice):
            import webbrowser

            webbrowser.open(choice.pkg.url)

    if skip in ic.choices:

        @bindings.add('s', eager=True)
        def skip_question(event: KeyPressEvent):
            ic.pointed_at = -1
            set_answer(event)

    @bindings.add(Keys.Any)
    def default(event: KeyPressEvent):
        # Disallow inserting other text
        pass

    layout = create_inquirer_layout(ic, get_prompt_tokens, **inquirer_kwargs)
    return Question(Application(layout=layout, key_bindings=bindings, style=qstyle))


def ask(question: Question) -> Any:
    return asyncio.run(question.application.run_async())


def _format_mb(value: int):
    return f'{value / 2 ** 20:.1f}'


class _DownloadProgress(formatters.Progress):
    template = '<current>{current:>3}</current>/<total>{total:>3}</total>MB'

    def format(self, progress_bar: ProgressBar, progress: Any, width: int):
        return HTML(self.template).format(
            current=_format_mb(progress.items_completed),
            total=_format_mb(progress.total) if progress.total is not None else '?',
        )


def make_progress_bar() -> ProgressBar:
    "``ProgressBar`` with download progress expressed in megabytes."
    return ProgressBar(
        formatters=[
            formatters.Label(),
            formatters.Text(' '),
            formatters.Percentage(),
            formatters.Text(' '),
            formatters.Bar(),
            formatters.Text(' '),
            _DownloadProgress(),
            formatters.Text(' '),
            formatters.Text('eta [', style='class:time-left'),
            formatters.TimeLeft(),
            formatters.Text(']', style='class:time-left'),
            formatters.Text(' '),
        ],
    )



from setuptools import setup

setup(
    name='instawow_test_plugin',
    py_modules=['instawow_test_plugin'],
    entry_points={'instawow.plugins': ['instawow_test_plugin = instawow_test_plugin']},
)



from __future__ import annotations

import pytest
from yarl import URL

from instawow.manager import Manager
from instawow.models import Pkg
from instawow.resolvers import Defn
from instawow.wa_updater import WaCompanionBuilder, WeakAura, WeakAuras


@pytest.fixture
def wa_saved_vars(iw_manager: Manager):
    saved_vars = (
        iw_manager.config.addon_dir.parents[1] / 'WTF' / 'Account' / 'test' / 'SavedVariables'
    )
    saved_vars.mkdir(parents=True)
    (saved_vars / WeakAuras.filename).write_text(
        '''\
WeakAurasSaved = {
    ["displays"] = {
        ["Foo"] = {
            ["bar"] = "baz",
        },
    },
}
'''
    )


@pytest.fixture
def builder(iw_manager: Manager):
    yield WaCompanionBuilder(iw_manager)


def test_can_parse_empty_displays_table(builder: WaCompanionBuilder):
    assert (
        builder.extract_auras(
            WeakAuras,
            '''\
WeakAurasSaved = {
    ["displays"] = {
    },
}
''',
        ).__root__
        == {}
    )


def test_urlless_display_is_discarded(builder: WaCompanionBuilder):
    assert (
        builder.extract_auras(
            WeakAuras,
            '''\
WeakAurasSaved = {
    ["displays"] = {
        ["Foo"] = {
            ["bar"] = "baz",
        },
    },
}
''',
        ).__root__
        == {}
    )


def test_can_parse_minimal_wago_display(builder: WaCompanionBuilder):
    aura = WeakAura(
        id='foo',
        uid='foo',
        parent=None,
        url=URL('https://wago.io/foo/1'),
        version=1,
    )
    assert (
        builder.extract_auras(
            WeakAuras,
            '''\
WeakAurasSaved = {
    ["displays"] = {
        ["Foo"] = {
            ["bar"] = "baz",
            ["url"] = "https://wago.io/foo/1",
            ["version"] = 1,
            ["id"] = "foo",
            ["uid"] = "foo",
        },
    },
}
''',
        ).__root__
        == {'foo': [aura]}
    )


def test_url_host_not_wago_display_is_discarded(builder: WaCompanionBuilder):
    assert (
        builder.extract_auras(
            WeakAuras,
            '''\
WeakAurasSaved = {
    ["displays"] = {
        ["Foo"] = {
            ["bar"] = "baz",
            ["url"] = "https://wafo.io/foo/1",
            ["version"] = 1,
            ["id"] = "foo",
            ["uid"] = "foo",
        },
    },
}
''',
        ).__root__
        == {}
    )


def test_can_build_addon_with_empty_seq(builder: WaCompanionBuilder):
    builder._generate_addon([])


@pytest.mark.asyncio
async def test_can_build_addon_with_mock_saved_vars(
    builder: WaCompanionBuilder, wa_saved_vars: None
):
    await builder.build()


def test_build_is_reproducible(builder: WaCompanionBuilder):
    builder._generate_addon([])
    checksum = builder._checksum()
    builder._generate_addon([])
    assert checksum == builder._checksum()


def test_changelog_is_generated(builder: WaCompanionBuilder):
    builder._generate_addon([])
    assert builder.changelog_path.read_text() == 'n/a'


@pytest.mark.asyncio
async def test_can_resolve_wa_companion_pkg(builder: WaCompanionBuilder):
    await builder.build()
    defn = Defn('instawow', 'weakauras-companion')
    resolve_results = await builder.manager.resolve([defn])
    assert type(resolve_results[defn]) is Pkg


@pytest.mark.asyncio
async def test_can_resolve_wa_companion_autoupdate_pkg(builder: WaCompanionBuilder):
    defn = Defn('instawow', 'weakauras-companion-autoupdate')
    resolve_results = await builder.manager.resolve([defn])
    assert type(resolve_results[defn]) is Pkg



from __future__ import annotations

import aiohttp
import pytest

from instawow.github_auth import get_codes, poll_for_access_token


@pytest.mark.asyncio
async def test_github_oauth_flow(iw_web_client: aiohttp.ClientSession):
    codes = await get_codes(iw_web_client)
    access_token = await poll_for_access_token(
        iw_web_client, codes['device_code'], codes['interval']
    )
    assert access_token == 'gho_16C7e42F292c6912E7710c838347Ae178B4a'



from __future__ import annotations

from collections.abc import Iterable
from functools import lru_cache
from typing import Any

import click
import pluggy

from . import resolvers

_project_name = __package__

hookspec = pluggy.HookspecMarker(_project_name)
hookimpl = pluggy.HookimplMarker(_project_name)


class InstawowPlugin:
    "The plug-in interface."

    @hookspec
    def instawow_add_commands(self) -> Iterable[click.Command]:
        "Additional commands to register with ``click``."
        ...

    @hookspec
    def instawow_add_resolvers(self) -> Iterable[resolvers.Resolver]:
        "Additional resolvers to load."
        ...


@lru_cache(maxsize=None)
def load_plugins() -> Any:
    plugin_manager = pluggy.PluginManager(_project_name)
    plugin_manager.add_hookspecs(InstawowPlugin)
    plugin_manager.load_setuptools_entrypoints(f'{_project_name}.plugins')
    return plugin_manager.hook



from __future__ import annotations

from collections.abc import Mapping, Sequence
from datetime import datetime
import json
import typing
from typing import Any

from pydantic import BaseModel

from . import manager
from .config import Flavour
from .utils import bucketise, cached_property, normalise_names


class _BaseCatalogueSameAs(BaseModel):
    source: str
    id: str


class BaseCatatalogueEntry(BaseModel):
    source: str
    id: str
    slug: str = ''
    name: str
    url: str
    game_flavours: typing.Set[Flavour]
    download_count: int
    last_updated: datetime
    folders: typing.List[typing.Set[str]] = []
    same_as: typing.List[_BaseCatalogueSameAs] = []


class _SerialisedClassNameModel(BaseModel):
    def dict(self, *args: object, **kwargs: object):
        dict_ = super().dict(*args, **kwargs)
        dict_['_type'] = self.__class__.__name__
        return dict_


class _CatalogueSameAs(_SerialisedClassNameModel, _BaseCatalogueSameAs):
    pass


class CatalogueEntry(_SerialisedClassNameModel, BaseCatatalogueEntry):
    same_as: typing.List[_CatalogueSameAs] = []  # type: ignore
    normalised_name: str
    derived_download_score: float


class BaseCatalogue(BaseModel, json_encoders={set: sorted}):
    version = 5
    entries: typing.List[BaseCatatalogueEntry]

    @classmethod
    async def collate(cls, start_date: datetime | None):
        async with manager.init_web_client() as web_client:
            entries = [e for r in manager.Manager.RESOLVERS async for e in r.catalogue(web_client)]
            if start_date is not None:
                entries = [e for e in entries if e.last_updated >= start_date]
            return cls(entries=entries)


class Catalogue(BaseModel, keep_untouched=(cached_property,)):
    version = 2
    entries: typing.List[CatalogueEntry]
    curse_slugs: typing.Dict[str, str]

    @classmethod
    def from_base_catalogue(
        cls, base_catalogue_values: Sequence[object], start_date: datetime | None
    ):
        normalise = normalise_names('')

        base_entries = BaseCatalogue.parse_obj(base_catalogue_values).entries
        if start_date is not None:
            base_entries = [e for e in base_entries if e.last_updated >= start_date]

        most_downloads_per_source = {
            s: max(e.download_count for e in i)
            for s, i in bucketise(base_entries, key=lambda e: e.source).items()
        }
        same_as_from_github = {
            (s.source, s.id): [
                _CatalogueSameAs(source=e.source, id=e.id),
                *(i for i in e.same_as if i.source != s.source),
            ]
            for e in base_entries
            if e.source == 'github' and e.same_as
            for s in e.same_as
        }
        entries = [
            CatalogueEntry.parse_obj(
                {
                    **e.__dict__,
                    'same_as': e.same_as
                    if e.source == 'github'
                    else (same_as_from_github.get((e.source, e.id)) or e.same_as),
                    'normalised_name': normalise(e.name),
                    'derived_download_score': 0
                    if e.source == 'github'
                    else e.download_count / most_downloads_per_source[e.source],
                }
            )
            for e in base_entries
        ]
        return cls(
            entries=entries,
            curse_slugs={e.slug: e.id for e in entries if e.source == 'curse'},
        )

    @classmethod
    def from_cache(cls, raw_catalogue: bytes):
        def parse_model_obj(values: Mapping[str, Any]):
            if '_type' not in values:
                return values
            elif values['_type'] == '_CatalogueSameAs':
                return _CatalogueSameAs.construct(**values)
            elif values['_type'] == 'CatalogueEntry':
                return CatalogueEntry.construct(
                    **{
                        **values,
                        'game_flavours': {Flavour(f) for f in values['game_flavours']},
                        'last_updated': datetime.fromisoformat(values['last_updated']),
                        'folders': [set(f) for f in values['folders']],
                    }
                )
            else:
                raise ValueError('Unknown type', values['_type'])

        return cls.construct(**json.loads(raw_catalogue, object_hook=parse_model_obj))

    @cached_property
    def keyed_entries(self) -> dict[tuple[str, str], CatalogueEntry]:
        return {(e.source, e.id): e for e in self.entries}



# pyright: reportUnknownMemberType=false

from __future__ import annotations

import asyncio
from collections.abc import (
    Awaitable,
    Callable,
    Collection,
    Generator,
    Iterable,
    Iterator,
    Sequence,
    Set,
)
from contextlib import contextmanager
from datetime import datetime, timezone
from functools import partial, wraps
from itertools import chain, repeat
from pathlib import Path
import textwrap
from typing import Any, NoReturn, TypeVar, overload

import click
from loguru import logger

from . import __version__, _deferred_types, db
from . import manager as _manager
from . import models
from . import results as R
from .common import Flavour, Strategy
from .config import Config, GlobalConfig, setup_logging
from .plugins import load_plugins
from .resolvers import ChangelogFormat, Defn
from .utils import StrEnum, all_eq, cached_property, evolve_model_obj, gather, tabulate, uniq

_T = TypeVar('_T')
_F = TypeVar('_F', bound='Callable[..., object]')


class Report:
    SUCCESS_SYMBOL = click.style('✓', fg='green')
    FAILURE_SYMBOL = click.style('✗', fg='red')
    WARNING_SYMBOL = click.style('!', fg='blue')

    def __init__(
        self,
        results: Iterable[tuple[Defn, R.ManagerResult]],
        filter_fn: Callable[[R.ManagerResult], bool] = lambda _: True,
    ) -> None:
        self.results = list(results)
        self.filter_fn = filter_fn

    @property
    def exit_code(self) -> int:
        return any(
            isinstance(r, (R.ManagerError, R.InternalError)) and self.filter_fn(r)
            for _, r in self.results
        )

    @classmethod
    def _result_type_to_symbol(cls, result: R.ManagerResult) -> str:
        if isinstance(result, R.InternalError):
            return cls.WARNING_SYMBOL
        elif isinstance(result, R.ManagerError):
            return cls.FAILURE_SYMBOL
        else:
            return cls.SUCCESS_SYMBOL

    def __str__(self) -> str:
        return '\n'.join(
            f'{self._result_type_to_symbol(r)} {click.style(a.to_urn(), bold=True)}\n'
            f'{textwrap.fill(r.message, initial_indent=" " * 2, subsequent_indent=" " * 4)}'
            for a, r in self.results
            if self.filter_fn(r)
        )

    def generate(self) -> None:
        manager_wrapper: ManagerWrapper | None = click.get_current_context().obj
        if manager_wrapper and manager_wrapper.manager.config.global_config.auto_update_check:
            outdated, new_version = run_with_progress(_manager.is_outdated())
            if outdated:
                click.echo(f'{self.WARNING_SYMBOL} instawow-{new_version} is available')

        report = str(self)
        if report:
            click.echo(report)

    def generate_and_exit(self) -> NoReturn:
        self.generate()
        ctx: click.Context = click.get_current_context()
        ctx.exit(self.exit_code)


def _init_cli_web_client(
    progress_bar: _deferred_types.prompt_toolkit.shortcuts.ProgressBar,
    tickers: set[asyncio.Task[None]],
) -> _deferred_types.aiohttp.ClientSession:
    from aiohttp import TraceConfig, hdrs
    from prompt_toolkit.shortcuts import ProgressBarCounter

    TICK_INTERVAL = 0.1

    async def do_on_request_end(
        client_session: _deferred_types.aiohttp.ClientSession,
        trace_config_ctx: Any,
        params: _deferred_types.aiohttp.TraceRequestEndParams,
    ):
        trace_request_ctx: _manager.TraceRequestCtx = trace_config_ctx.trace_request_ctx
        if trace_request_ctx:
            response = params.response
            label = (
                'Downloading '
                + Defn(trace_request_ctx['pkg'].source, trace_request_ctx['pkg'].slug).to_urn()
                if trace_request_ctx['report_progress'] == 'pkg_download'
                else trace_request_ctx['label']
            )
            # The encoded size is not exposed in the aiohttp streaming API
            # so we cannot display progress when the payload is encoded.
            # When ``None`` the progress bar is "indeterminate".
            total = None if hdrs.CONTENT_ENCODING in response.headers else response.content_length

            async def ticker():
                counter: ProgressBarCounter[None] = ProgressBarCounter(
                    progress_bar=progress_bar, label=label, total=total
                )
                progress_bar.counters.append(counter)
                try:
                    while not response.content.is_eof():
                        counter.items_completed = response.content.total_bytes
                        progress_bar.invalidate()
                        await asyncio.sleep(TICK_INTERVAL)
                finally:
                    progress_bar.counters.remove(counter)

            tickers.add(asyncio.create_task(ticker()))

    trace_config = TraceConfig()
    trace_config.on_request_end.append(do_on_request_end)
    trace_config.freeze()
    return _manager.init_web_client(trace_configs=[trace_config])


@contextmanager
def _cancel_tickers(progress_bar: _deferred_types.prompt_toolkit.shortcuts.ProgressBar):
    tickers: set[asyncio.Task[None]] = set()
    try:
        yield tickers
    finally:
        for ticker in tickers:
            ticker.cancel()
        progress_bar.invalidate()


def run_with_progress(awaitable: Awaitable[_T]) -> _T:
    from .prompts import make_progress_bar

    async def run():
        with make_progress_bar() as progress_bar, _cancel_tickers(progress_bar) as tickers:
            async with _init_cli_web_client(progress_bar, tickers) as web_client:
                _manager.Manager.contextualise(web_client=web_client)
                return await awaitable

    return asyncio.run(run())


def _override_asyncio_loop_policy():
    # The proactor event loop which became the default loop on Windows
    # in Python 3.8 is causing issues with aiohttp.
    # See https://github.com/aio-libs/aiohttp/issues/4324
    policy = getattr(asyncio, 'WindowsSelectorEventLoopPolicy', None)
    if policy:
        asyncio.set_event_loop_policy(policy())


def _apply_patches():
    _override_asyncio_loop_policy()


class ManagerWrapper:
    def __init__(self, ctx: click.Context) -> None:
        self.ctx = ctx

    @cached_property
    def manager(self) -> _manager.Manager:
        try:
            config = Config.read(None, self.ctx.params['profile']).ensure_dirs()
        except FileNotFoundError:
            config = self.ctx.invoke(configure)

        setup_logging(
            config.logging_dir, self.ctx.params['log_level'], self.ctx.params['log_to_stderr']
        )
        manager, close_db_conn = _manager.Manager.from_config(config)
        self.ctx.call_on_close(close_db_conn)

        return manager

    @staticmethod
    def pass_manager(fn: _F) -> _F:
        @wraps(fn)
        def wrapper(*args: object, **kwargs: object):
            return fn(click.get_current_context().obj.manager, *args, **kwargs)

        return wrapper  # type: ignore


def _with_manager(fn: Callable[..., object]):
    def wrapper(ctx: click.Context, __: click.Parameter, value: object):
        return fn(ctx.obj.manager, value)

    return wrapper


class StrEnumParam(click.Choice):
    def __init__(
        self,
        choice_enum: type[StrEnum],
        excludes: Set[StrEnum] = frozenset(),
        case_sensitive: bool = True,
    ) -> None:
        self.choice_enum = choice_enum
        super().__init__(
            choices=[c for c in choice_enum if c not in excludes],
            case_sensitive=case_sensitive,
        )

    def convert(
        self, value: str, param: click.Parameter | None, ctx: click.Context | None
    ) -> StrEnum:
        return self.choice_enum(super().convert(value, param, ctx))


def _register_plugin_commands(group: click.Group):
    plugin_hook = load_plugins()
    additional_commands = (c for g in plugin_hook.instawow_add_commands() for c in g)
    for command in additional_commands:
        group.add_command(command)
    return group


def _set_log_level(_: click.Context, __: click.Parameter, value: bool):
    return 'DEBUG' if value else 'INFO'


@_register_plugin_commands
@click.group(context_settings={'help_option_names': ('-h', '--help')})
@click.version_option(__version__, prog_name=__package__)
@click.option(
    '--debug',
    'log_level',
    is_flag=True,
    default=False,
    callback=_set_log_level,
    help='Log more things.',
)
@click.option(
    '--log-to-stderr',
    is_flag=True,
    default=False,
    help='Log to stderr for development.',
    hidden=True,
)
@click.option(
    '--profile',
    '-p',
    default='__default__',
    help='Activate the specified profile.',
)
@click.pass_context
def cli(ctx: click.Context, log_level: str, log_to_stderr: bool, profile: str) -> None:
    "Add-on manager for World of Warcraft."
    _apply_patches()
    ctx.obj = ManagerWrapper(ctx)


main = logger.catch(reraise=True)(cli)


@overload
def parse_into_defn(manager: _manager.Manager, value: str, *, raise_invalid: bool = True) -> Defn:
    ...


@overload
def parse_into_defn(
    manager: _manager.Manager, value: list[str], *, raise_invalid: bool = True
) -> list[Defn]:
    ...


@overload
def parse_into_defn(manager: _manager.Manager, value: None, *, raise_invalid: bool = True) -> None:
    ...


def parse_into_defn(
    manager: _manager.Manager, value: str | list[str] | None, *, raise_invalid: bool = True
) -> Defn | list[Defn] | None:
    if value is None:
        return None

    if not isinstance(value, str):
        defns = (parse_into_defn(manager, v, raise_invalid=raise_invalid) for v in value)
        return uniq(defns)

    pair = manager.pair_uri(value)
    if not pair:
        if raise_invalid:
            raise click.BadParameter(value)

        pair = ('*', value)
    return Defn(*pair)


def parse_into_defn_with_strategy(
    manager: _manager.Manager, value: Sequence[tuple[Strategy, str]]
) -> Iterator[Defn]:
    defns = parse_into_defn(manager, [d for _, d in value])
    return map(lambda d, s: evolve_model_obj(d, strategy=s), defns, (s for s, _ in value))


def parse_into_defn_with_version(
    manager: _manager.Manager, value: Sequence[tuple[str, str]]
) -> Iterator[Defn]:
    defns = parse_into_defn(manager, [d for _, d in value])
    return map(Defn.with_version, defns, (v for v, _ in value))


def _combine_addons(
    fn: Callable[[_manager.Manager, _T], Iterable[Defn]],
    ctx: click.Context,
    _: click.Parameter,
    value: _T,
):
    addons: list[Defn] = ctx.params.setdefault('addons', [])
    if value:
        addons.extend(fn(ctx.obj.manager, value))


_EXCLUDED_STRATEGIES = frozenset({Strategy.default, Strategy.version})


@cli.command()
@click.argument(
    'addons', nargs=-1, callback=partial(_combine_addons, parse_into_defn), expose_value=False
)
@click.option(
    '--with-strategy',
    '-s',
    multiple=True,
    type=(StrEnumParam(Strategy, _EXCLUDED_STRATEGIES), str),
    expose_value=False,
    callback=partial(_combine_addons, parse_into_defn_with_strategy),
    metavar='<STRATEGY ADDON>...',
    help='A strategy followed by an add-on definition.  '
    f'One of: {", ".join(s for s in Strategy if s not in _EXCLUDED_STRATEGIES)}.',
)
@click.option(
    '--version',
    multiple=True,
    type=(str, str),
    expose_value=False,
    callback=partial(_combine_addons, parse_into_defn_with_version),
    metavar='<VERSION ADDON>...',
    help='A version followed by an add-on definition.',
)
@click.option('--replace', is_flag=True, default=False, help='Replace unreconciled add-ons.')
@ManagerWrapper.pass_manager
def install(manager: _manager.Manager, addons: Sequence[Defn], replace: bool) -> None:
    "Install add-ons."
    if not addons:
        raise click.UsageError(
            'You must provide at least one of "ADDONS", "--with-strategy" or "--version"'
        )

    results = run_with_progress(manager.install(addons, replace))
    Report(results.items()).generate_and_exit()


@cli.command()
@click.argument('addons', nargs=-1, callback=_with_manager(parse_into_defn))
@ManagerWrapper.pass_manager
def update(manager: _manager.Manager, addons: Sequence[Defn]) -> None:
    "Update installed add-ons."
    import sqlalchemy as sa

    def filter_results(result: R.ManagerResult):
        # Hide packages from output if they are up to date
        # and ``update`` was invoked without args,
        # provided that they are not pinned
        if addons or not isinstance(result, R.PkgUpToDate):
            return True
        else:
            return result.is_pinned

    update_defns = addons or [
        Defn(**v)
        for v in manager.database.execute(
            sa.select(
                db.pkg.c.source,
                db.pkg.c.id,
                db.pkg.c.slug.label('alias'),
                db.pkg.c.version,
                db.pkg_options.c.strategy,
            ).join(db.pkg_options)
        )
        .mappings()
        .all()
    ]
    results = run_with_progress(manager.update(update_defns, False))
    Report(results.items(), filter_results).generate_and_exit()


@cli.command()
@click.argument('addons', nargs=-1, required=True, callback=_with_manager(parse_into_defn))
@click.option(
    '--keep-folders',
    is_flag=True,
    default=False,
    help="Do not delete the add-on folders.",
)
@ManagerWrapper.pass_manager
def remove(manager: _manager.Manager, addons: Sequence[Defn], keep_folders: bool) -> None:
    "Remove add-ons."
    results = run_with_progress(manager.remove(addons, keep_folders))
    Report(results.items()).generate_and_exit()


@cli.command()
@click.argument('addon', callback=_with_manager(parse_into_defn))
@click.option(
    '--version',
    help='Version to roll back to.',
)
@click.option(
    '--undo',
    is_flag=True,
    default=False,
    help='Undo rollback by reinstalling an add-on using the default strategy.',
)
@ManagerWrapper.pass_manager
def rollback(manager: _manager.Manager, addon: Defn, version: str | None, undo: bool) -> None:
    "Roll an add-on back to an older version."
    from .prompts import Choice, ask, select

    if version and undo:
        raise click.UsageError('Cannot use "--version" with "--undo"')

    pkg = manager.get_pkg(addon)
    if not pkg:
        report: Report = Report([(addon, R.PkgNotInstalled())])
        report.generate_and_exit()

    if Strategy.version not in manager.resolvers[pkg.source].strategies:
        Report([(addon, R.PkgStrategyUnsupported(Strategy.version))]).generate_and_exit()

    if undo:
        Report(run_with_progress(manager.update([addon], True)).items()).generate_and_exit()

    reconstructed_defn = Defn.from_pkg(pkg)
    if version:
        selection = version
    else:
        versions = pkg.logged_versions
        if len(versions) <= 1:
            Report(
                [(addon, R.PkgFileUnavailable('cannot find older versions'))]
            ).generate_and_exit()

        choices = [
            Choice(
                [('', v.version)],
                value=v.version,
                disabled='installed version' if v.version == pkg.version else None,
            )
            for v in versions
        ]
        selection = ask(
            select(f'Select version of {reconstructed_defn.to_urn()} for rollback', choices)
        )

    Report(
        run_with_progress(
            manager.update([reconstructed_defn.with_version(selection)], True)
        ).items()
    ).generate_and_exit()


@cli.command()
@click.option('--auto', '-a', is_flag=True, default=False, help='Do not ask for user input.')
@click.option(
    '--installed',
    'rereconcile',
    is_flag=True,
    default=False,
    help='Re-reconcile installed add-ons.',
)
@click.option(
    '--list-unreconciled', is_flag=True, default=False, help='List unreconciled add-ons and exit.'
)
@ManagerWrapper.pass_manager
def reconcile(
    manager: _manager.Manager, auto: bool, rereconcile: bool, list_unreconciled: bool
) -> None:
    "Reconcile pre-installed add-ons."
    from .matchers import (
        AddonFolder,
        get_unreconciled_folders,
        match_addon_names_with_folder_names,
        match_folder_name_subsets,
        match_toc_source_ids,
    )
    from .prompts import PkgChoice, ask, confirm, select, skip

    def construct_choice(pkg: models.Pkg, highlight_version: bool, disabled: bool):
        defn = Defn.from_pkg(pkg)
        return PkgChoice(
            [
                ('', f'{defn.to_urn()}=='),
                ('class:highlight-sub' if highlight_version else '', pkg.version),
            ],
            pkg=pkg,
            value=defn,
            disabled=disabled,
        )

    def gather_selections(
        groups: Sequence[tuple[Any, Sequence[Defn]]],
        selector: Callable[[Any, Sequence[models.Pkg]], Defn | None],
    ):
        results = run_with_progress(manager.resolve(uniq(d for _, b in groups for d in b)))
        for addons_or_pkg, defns in groups:
            shortlist = [r for d in defns for r in (results[d],) if isinstance(r, models.Pkg)]
            if shortlist:
                selection = selector(addons_or_pkg, shortlist)
                yield selection

    if rereconcile:
        if auto:
            raise click.UsageError('Cannot use "--auto" with "--installed"')

        import sqlalchemy as sa

        def prompt_reconciled(installed_pkg: models.Pkg, pkgs: Sequence[models.Pkg]):
            highlight_version = not all_eq(i.version for i in (installed_pkg, *pkgs))
            choices = [
                construct_choice(installed_pkg, highlight_version, True),
                *(construct_choice(p, highlight_version, False) for p in pkgs),
                skip,
            ]
            selection = ask(select(installed_pkg.name, choices))
            return selection or None

        installed_pkgs = (
            models.Pkg.from_row_mapping(manager.database, p)
            for p in manager.database.execute(sa.select(db.pkg)).mappings().all()
        )
        groups = list(run_with_progress(manager.find_equivalent_pkg_defns(installed_pkgs)).items())
        selections = [
            (p, s) for (p, _), s in zip(groups, gather_selections(groups, prompt_reconciled)) if s
        ]
        if selections and ask(confirm('Install selected add-ons?')):
            Report(
                run_with_progress(
                    manager.remove([Defn.from_pkg(p) for p, _ in selections], False),
                ).items()
            ).generate()
            Report(
                run_with_progress(
                    manager.install([s for _, s in selections], False),
                ).items()
            ).generate_and_exit()

        return

    preamble = textwrap.dedent(
        '''\
        Use the arrow keys to navigate, <o> to open an add-on in your browser,
        enter to make a selection and <s> to skip to the next item.

        Versions that differ from the installed version or differ between
        choices are highlighted in purple.

        The reconciler will perform three passes in decreasing order of accuracy,
        looking to match source IDs and add-on names in TOC files, and folders.

        Selected add-ons will be reinstalled.

        You can also run `reconcile` in promptless mode by passing
        the `--auto` flag.  In this mode, add-ons will be reconciled
        without user input.
        '''
    )

    def prompt_unreconciled(addons: Sequence[AddonFolder], pkgs: Sequence[models.Pkg]):
        def combine_names():
            return textwrap.shorten(', '.join(a.name for a in addons), 60)

        # Highlight version if there's multiple of them
        highlight_version = not all_eq(i.version for i in chain(addons, pkgs))
        choices = [
            *(construct_choice(p, highlight_version, False) for p in pkgs),
            skip,
        ]
        selection = ask(select(f'{combine_names()} [{addons[0].version or "?"}]', choices))
        return selection or None

    def pick_first(addons: Sequence[AddonFolder], pkgs: Sequence[models.Pkg]):
        return Defn.from_pkg(pkgs[0])

    def match_all(
        selector: Callable[[Any, Sequence[models.Pkg]], Defn | None]
    ) -> Generator[list[Defn], frozenset[AddonFolder], None]:
        # Match in order of increasing heuristicitivenessitude
        for fn in [
            match_toc_source_ids,
            match_folder_name_subsets,
            match_addon_names_with_folder_names,
        ]:
            groups = run_with_progress(fn(manager, (yield [])))
            yield list(filter(None, gather_selections(groups, selector)))

    leftovers = get_unreconciled_folders(manager)
    if list_unreconciled:
        table_rows = [('unreconciled',), *((f.name,) for f in sorted(leftovers))]
        click.echo(tabulate(table_rows))
        return
    elif not leftovers:
        click.echo('No add-ons left to reconcile.')
        return

    if not auto:
        click.echo(preamble)

    matcher = match_all(pick_first if auto else prompt_unreconciled)
    for _ in matcher:  # Skip over consumer yields
        selections = matcher.send(leftovers)
        if selections and (auto or ask(confirm('Install selected add-ons?'))):
            results = run_with_progress(manager.install(selections, True))
            Report(results.items()).generate()

        leftovers = get_unreconciled_folders(manager)

    if leftovers:
        click.echo()
        table_rows = [('unreconciled',), *((f.name,) for f in sorted(leftovers))]
        click.echo(tabulate(table_rows))


def _concat_search_terms(_: click.Context, __: click.Parameter, value: tuple[str, ...]):
    return ' '.join(value)


def _parse_iso_date_into_datetime(_: click.Context, __: click.Parameter, value: str | None):
    if value is not None:
        return datetime.strptime(value, '%Y-%m-%d').replace(tzinfo=timezone.utc)


@cli.command()
@click.argument('search-terms', nargs=-1, required=True, callback=_concat_search_terms)
@click.option(
    '--limit',
    '-l',
    default=10,
    type=click.IntRange(1, 20, clamp=True),
    help='A number to limit results to.',
)
@click.option(
    '--start-date',
    callback=_parse_iso_date_into_datetime,
    help='Omit results before this date.',
    metavar='YYYY-MM-DD',
)
@click.option(
    '--source',
    'sources',
    multiple=True,
    help='A source to search in.  Repeatable.',
)
@click.pass_context
def search(
    ctx: click.Context,
    search_terms: str,
    limit: int,
    sources: Sequence[str],
    start_date: datetime | None,
) -> None:
    "Search for add-ons to install."
    from .prompts import PkgChoice, ask, checkbox, confirm

    manager: _manager.Manager = ctx.obj.manager

    entries = run_with_progress(
        manager.search(search_terms, limit, frozenset(sources), start_date)
    )
    defns = [Defn(e.source, e.id) for e in entries]
    pkgs = (
        (evolve_model_obj(d, alias=r.slug), r)
        for d, r in run_with_progress(manager.resolve(defns)).items()
        if isinstance(r, models.Pkg)
    )
    choices = [PkgChoice(f'{p.name}  ({d.to_urn()}=={p.version})', d, pkg=p) for d, p in pkgs]
    if choices:
        selections: list[Defn] = ask(checkbox('Select add-ons to install', choices=choices))
        if selections and ask(confirm('Install selected add-ons?')):
            ctx.invoke(install, addons=selections)
    else:
        click.echo('No results found.')


class _ListFormats(StrEnum):
    simple = 'simple'
    detailed = 'detailed'
    json = 'json'


@cli.command('list')
@click.argument(
    'addons', nargs=-1, callback=_with_manager(partial(parse_into_defn, raise_invalid=False))
)
@click.option(
    '--format',
    '-f',
    'output_format',
    type=StrEnumParam(_ListFormats),
    default=_ListFormats.simple,
    show_default=True,
    help='Change the output format.',
)
@ManagerWrapper.pass_manager
def list_installed(
    manager: _manager.Manager, addons: Sequence[Defn], output_format: _ListFormats
) -> None:
    "List installed add-ons."
    import sqlalchemy as sa

    def format_deps(pkg: models.Pkg):
        return (
            Defn(pkg.source, s or e.id).to_urn()
            for e in pkg.deps
            for s in (
                manager.database.execute(
                    sa.select(db.pkg.c.slug).filter_by(source=pkg.source, id=e.id)
                ).scalar_one_or_none(),
            )
        )

    def row_mappings_to_pkgs():
        return map(models.Pkg.from_row_mapping, repeat(manager.database), pkg_mappings)

    pkg_mappings = (
        manager.database.execute(
            sa.select(db.pkg)
            .filter(
                sa.or_(
                    *(
                        db.pkg.c.slug.contains(d.alias)
                        if d.source == '*'
                        else (db.pkg.c.source == d.source)
                        & ((db.pkg.c.id == d.alias) | (db.pkg.c.slug == d.alias))
                        for d in addons
                    )
                )
                if addons
                else True
            )
            .order_by(db.pkg.c.source, db.pkg.c.name)
        )
        .mappings()
        .all()
    )

    if output_format is _ListFormats.json:
        click.echo(models.PkgList.parse_obj(list(row_mappings_to_pkgs())).json(indent=2))

    elif output_format is _ListFormats.detailed:
        formatter = click.HelpFormatter(max_width=99)
        for pkg in row_mappings_to_pkgs():
            with formatter.section(Defn.from_pkg(pkg).to_urn()):
                formatter.write_dl(
                    [
                        ('name', pkg.name),
                        ('description', textwrap.shorten(pkg.description, 280)),
                        ('url', pkg.url),
                        ('version', pkg.version),
                        ('date_published', pkg.date_published.isoformat(' ', 'minutes')[:-6]),
                        ('folders', ', '.join(f.name for f in pkg.folders)),
                        ('deps', ', '.join(format_deps(pkg))),
                        ('options.strategy', pkg.options.strategy),
                    ]
                )
        click.echo(formatter.getvalue(), nl=False)

    else:
        click.echo(
            ''.join(f'{Defn(p["source"], p["slug"]).to_urn()}\n' for p in pkg_mappings),
            nl=False,
        )


@cli.command(hidden=True)
@click.argument('addon', callback=_with_manager(partial(parse_into_defn, raise_invalid=False)))
@click.pass_context
def info(ctx: click.Context, addon: Defn) -> None:
    "Alias of `list -f detailed`."
    ctx.invoke(list_installed, addons=(addon,), output_format=_ListFormats.detailed)


@cli.command()
@click.argument('addon', callback=_with_manager(partial(parse_into_defn, raise_invalid=False)))
@ManagerWrapper.pass_manager
def reveal(manager: _manager.Manager, addon: Defn) -> None:
    "Bring an add-on up in your file manager."
    pkg = manager.get_pkg(addon, partial_match=True)
    if pkg:
        click.launch(str(manager.config.addon_dir / pkg.folders[0].name), locate=True)
    else:
        Report([(addon, R.PkgNotInstalled())]).generate_and_exit()


@cli.command()
@click.argument(
    'addon', callback=_with_manager(partial(parse_into_defn, raise_invalid=False)), required=False
)
@click.option(
    '--convert/--no-convert',
    default=True,
    show_default=True,
    help='Convert HTML and Markdown changelogs to plain text using pandoc.',
)
@ManagerWrapper.pass_manager
def view_changelog(manager: _manager.Manager, addon: Defn | None, convert: bool) -> None:
    """View the changelog of an installed add-on.

    If `addon` is not provided, displays the changelogs of all add-ons
    to have been installed within one minute of the last add-on.
    """

    def make_converter():
        import shutil

        pandoc = shutil.which('pandoc')
        if pandoc is None:

            def noop_convert(source: str, changelog: str):
                return changelog

            return noop_convert
        else:

            def real_convert(source: str, changelog: str):
                import subprocess

                changelog_format = manager.resolvers[source].changelog_format
                if changelog_format not in {ChangelogFormat.html, ChangelogFormat.markdown}:
                    return changelog
                else:
                    return subprocess.check_output(
                        [pandoc, '-f', changelog_format, '-t', 'plain'], input=changelog, text=True
                    )

            return real_convert

    if addon:
        pkg = manager.get_pkg(addon, partial_match=True)
        if pkg:
            changelog = run_with_progress(manager.get_changelog(pkg.changelog_url))
            if convert:
                changelog = make_converter()(pkg.source, changelog)
            click.echo_via_pager(changelog)
        else:
            Report([(addon, R.PkgNotInstalled())]).generate_and_exit()

    else:
        import sqlalchemy as sa

        last_installed_changelog_urls = manager.database.execute(
            sa.select(db.pkg.c.source, db.pkg.c.slug, db.pkg.c.changelog_url)
            .join(db.pkg_version_log)
            .filter(
                db.pkg_version_log.c.install_time
                >= sa.select(
                    sa.func.datetime(sa.func.max(db.pkg_version_log.c.install_time), '-1 minute')
                )
                .join(db.pkg)
                .scalar_subquery()
            )
        ).all()
        changelogs = run_with_progress(
            gather(manager.get_changelog(m.changelog_url) for m in last_installed_changelog_urls)
        )
        if convert:
            do_convert = make_converter()
            changelogs = (
                do_convert(m.source, c) for m, c in zip(last_installed_changelog_urls, changelogs)
            )
        click.echo_via_pager(
            '\n\n'.join(
                Defn(m.source, m.slug).to_urn() + ':\n' + textwrap.indent(c, '  ')
                for m, c in zip(last_installed_changelog_urls, changelogs)
            )
        )


def _show_active_config(ctx: click.Context, __: click.Parameter, value: bool):
    if value:
        click.echo(ctx.obj.manager.config.json(indent=2))
        ctx.exit()


async def _github_oauth_flow():
    from .github_auth import get_codes, poll_for_access_token

    async with _manager.init_web_client() as web_client:
        codes = await get_codes(web_client)
        click.echo(f'Navigate to {codes["verification_uri"]} and paste the code below:')
        click.echo(f'  {codes["user_code"]}')
        click.echo('Waiting...')
        access_token = await poll_for_access_token(
            web_client, codes['device_code'], codes['interval']
        )
        return access_token


class _EditableConfigOptions(StrEnum):
    addon_dir = 'addon_dir'
    game_flavour = 'game_flavour'
    auto_update_check = 'auto_update_check'
    github_access_token = 'access_tokens.github'


@cli.command()
@click.option(
    '--show-active',
    is_flag=True,
    default=False,
    expose_value=False,
    is_eager=True,
    callback=_show_active_config,
    help='Show the active configuration and exit.',
)
@click.argument(
    'config-options',
    nargs=-1,
    type=StrEnumParam(_EditableConfigOptions),
)
@click.pass_context
def configure(
    ctx: click.Context,
    config_options: Collection[_EditableConfigOptions],
) -> Config:
    "Configure instawow."
    from .prompts import PydanticValidator, ask, confirm, path, select

    if not config_options:
        config_options = {
            _EditableConfigOptions.addon_dir,
            _EditableConfigOptions.game_flavour,
            _EditableConfigOptions.github_access_token,
        }

    profile = ctx.find_root().params['profile']

    orig_global_config = GlobalConfig.read()
    global_config_values = orig_global_config.dict()
    try:
        profile_config_values = Config.read(orig_global_config, profile).dict(
            exclude={'global_config'}
        )
    except FileNotFoundError:
        profile_config_values = {
            'profile': profile,
        }

    addon_dir = None
    if _EditableConfigOptions.addon_dir in config_options:
        addon_dir = ask(
            path(
                'Add-on directory:',
                only_directories=True,
                validate=PydanticValidator(Config, 'addon_dir'),
            )
        )
        profile_config_values['addon_dir'] = addon_dir

    if _EditableConfigOptions.game_flavour in config_options:
        game_flavour = ask(
            select(
                'Game flavour:',
                choices=list(Flavour),
                initial_choice=Config.infer_flavour(addon_dir) if addon_dir is not None else None,
            )
        )
        profile_config_values['game_flavour'] = game_flavour

    if _EditableConfigOptions.auto_update_check in config_options:
        global_config_values['auto_update_check'] = ask(
            confirm('Periodically check for instawow updates?')
        )

    if _EditableConfigOptions.github_access_token in config_options and ask(
        confirm('Set up GitHub authentication?')
    ):
        github_access_token = run_with_progress(_github_oauth_flow())
        global_config_values['access_tokens']['github'] = github_access_token

    global_config = GlobalConfig(**global_config_values).write()
    config = Config(global_config=global_config, **profile_config_values).write()

    click.echo('Configuration written to:')
    click.echo(f'  {global_config.config_file}')
    click.echo(f'  {config.config_file}')

    return config


@cli.group('weakauras-companion')
def _weakauras_group() -> None:
    "Manage your WeakAuras."


@_weakauras_group.command('build')
@ManagerWrapper.pass_manager
def build_weakauras_companion(manager: _manager.Manager) -> None:
    "Build the WeakAuras Companion add-on."
    from .wa_updater import WaCompanionBuilder

    run_with_progress(WaCompanionBuilder(manager).build())


@_weakauras_group.command('list')
@ManagerWrapper.pass_manager
def list_installed_wago_auras(manager: _manager.Manager) -> None:
    "List WeakAuras installed from Wago."
    from .wa_updater import WaCompanionBuilder

    aura_groups = WaCompanionBuilder(manager).extract_installed_auras()
    installed_auras = sorted(
        (g.filename, textwrap.fill(a.id, width=30, max_lines=1), a.url)
        for g in aura_groups
        for v in g.__root__.values()
        for a in v
        if not a.parent
    )
    click.echo(tabulate([('in file', 'name', 'URL'), *installed_auras]))


@cli.command(hidden=True)
@click.option(
    '--start-date',
    callback=_parse_iso_date_into_datetime,
    help='Omit results before this date.',
    metavar='YYYY-MM-DD',
)
def generate_catalogue(start_date: datetime | None) -> None:
    "Generate the master catalogue."
    from .cataloguer import BaseCatalogue

    catalogue = asyncio.run(BaseCatalogue.collate(start_date))
    catalogue_path = Path(f'base-catalogue-v{catalogue.version}.json').resolve()
    catalogue_path.write_text(
        catalogue.json(indent=2),
        encoding='utf-8',
    )
    catalogue_path.with_suffix(f'.compact{catalogue_path.suffix}').write_text(
        catalogue.json(separators=(',', ':')),
        encoding='utf-8',
    )


@cli.command()
@click.pass_context
def gui(ctx: click.Context) -> None:
    "Fire up the GUI."
    from instawow_gui.app import InstawowApp

    global_config = GlobalConfig.read().ensure_dirs()
    dummy_jsonrpc_config = Config.construct(
        global_config=global_config, profile='__jsonrpc__'
    ).ensure_dirs()
    params = ctx.find_root().params
    setup_logging(dummy_jsonrpc_config.logging_dir, params['log_level'], params['log_to_stderr'])

    InstawowApp(version=__version__).main_loop()



from __future__ import annotations

from datetime import datetime, timezone
from enum import IntEnum

from sqlalchemy import (
    Column,
    DateTime,
    ForeignKeyConstraint,
    MetaData,
    String,
    Table,
    TypeDecorator,
    func,
    text,
)
from sqlalchemy.exc import OperationalError
from sqlalchemy.future import Engine


class TZDateTime(TypeDecorator[datetime]):
    impl = DateTime
    cache_ok = True

    def process_bind_param(self, value: datetime | None, dialect: object) -> datetime | None:
        if value is not None:
            if not value.tzinfo:
                raise TypeError('tzinfo is required')
            value = value.astimezone(timezone.utc).replace(tzinfo=None)
        return value

    def process_result_value(self, value: datetime | None, dialect: object) -> datetime | None:
        if value is not None:
            value = value.replace(tzinfo=timezone.utc)
        return value

    @property
    def python_type(self) -> type[datetime]:
        return datetime


metadata = MetaData()


pkg = Table(
    'pkg',
    metadata,
    Column('source', String, primary_key=True),
    Column('id', String, primary_key=True),
    Column('slug', String, nullable=False),
    Column('name', String, nullable=False),
    Column('description', String, nullable=False),
    Column('url', String, nullable=False),
    Column('download_url', String, nullable=False),
    Column('date_published', TZDateTime, nullable=False),
    Column('version', String, nullable=False),
    Column('changelog_url', String, nullable=False),
)

pkg_options = Table(
    'pkg_options',
    metadata,
    Column('strategy', String, nullable=False),
    Column('pkg_source', String, primary_key=True),
    Column('pkg_id', String, primary_key=True),
    ForeignKeyConstraint(
        ['pkg_source', 'pkg_id'],
        ['pkg.source', 'pkg.id'],
        name='fk_pkg_options_pkg_source_and_id',
    ),
)

pkg_folder = Table(
    'pkg_folder',
    metadata,
    Column('name', String, primary_key=True),
    Column('pkg_source', String, nullable=False),
    Column('pkg_id', String, nullable=False),
    ForeignKeyConstraint(
        ['pkg_source', 'pkg_id'],
        ['pkg.source', 'pkg.id'],
        name='fk_pkg_folder_pkg_source_and_id',
    ),
)

pkg_dep = Table(
    'pkg_dep',
    metadata,
    Column('id', String, primary_key=True),
    Column('pkg_source', String, primary_key=True),
    Column('pkg_id', String, primary_key=True),
    ForeignKeyConstraint(
        ['pkg_source', 'pkg_id'],
        ['pkg.source', 'pkg.id'],
        name='fk_pkg_dep_pkg_source_and_id',
    ),
)

pkg_version_log = Table(
    'pkg_version_log',
    metadata,
    Column('version', String, primary_key=True),
    Column('install_time', TZDateTime, nullable=False, server_default=func.now()),
    Column('pkg_source', String, primary_key=True),
    Column('pkg_id', String, primary_key=True),
    ForeignKeyConstraint(
        ['pkg_source', 'pkg_id'],
        ['pkg.source', 'pkg.id'],
        name='fk_pkg_version_log_pkg_source_and_id',
    ),
)


class DatabaseState(IntEnum):
    current = 0
    old = 1
    uninitialised = 2


def get_database_state(engine: Engine, revision: str) -> DatabaseState:
    with engine.connect() as connection:
        try:
            state = connection.execute(
                text(
                    'SELECT ifnull((SELECT 0 FROM alembic_version WHERE version_num == :revision), 1)',
                ),
                {'revision': revision},
            ).scalar()
        except OperationalError:
            state = connection.execute(
                text(
                    'SELECT ifnull((SELECT 1 FROM sqlite_master WHERE type = "table" LIMIT 1), 2)',
                )
            ).scalar()

    return DatabaseState(state)


def migrate_database(engine: Engine, revision: str) -> None:
    database_state = get_database_state(engine, revision)
    if database_state != DatabaseState.current:
        import alembic.command
        import alembic.config

        alembic_config = alembic.config.Config()
        alembic_config.set_main_option('script_location', f'{__package__}:migrations')
        alembic_config.set_main_option('sqlalchemy.url', str(engine.url))

        if database_state == DatabaseState.uninitialised:
            metadata.create_all(engine)
            alembic.command.stamp(alembic_config, revision)
        else:
            alembic.command.upgrade(alembic_config, revision)



"""
Drop the package table's ``file_id`` column.
"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'f3f9957de30c'
down_revision = '8f6ba74cfa82'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('pkg', schema=None) as batch_op:
        batch_op.drop_column('file_id')


def downgrade():
    with op.batch_alter_table('pkg', schema=None) as batch_op:
        batch_op.add_column(sa.Column('file_id', sa.VARCHAR(), nullable=False))



# This module is adapted from SLPP <https://github.com/SirAnthony/slpp>.
#
#   Copyright (c) 2010, 2011, 2012 SirAnthony <anthony at adsorbtion.org>
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.

from __future__ import annotations

from itertools import count, islice
from operator import eq
import re
import string
from typing import Any, Container

DIGITS = frozenset(string.digits)
HEXDIGITS = frozenset(string.hexdigits)
HEXDELIMS = frozenset('Xx')
EXPONENTS = frozenset('Ee')
WHITESPACE = frozenset(string.whitespace)
WHITESPACE_OR_CLOSING_SQ_BR = WHITESPACE | frozenset(']')
NEWLINE = frozenset('\r\n')

match_bare_word = re.compile(r'^[a-z_]\w*$', flags=re.IGNORECASE)


class ParseError(Exception):
    pass


class _Sentinel(str):
    pass


_sentinel = _Sentinel()


class _SLPP:
    def __init__(self, text: str):
        self._iter_text = iter(text)
        self._next()

    def _next(self):
        self.c = next(self._iter_text, _sentinel)

    def _next_eq(self, includes: Container[str]):
        if self.c not in includes:
            for c in self._iter_text:
                if c in includes:
                    self.c = c
                    break
            else:
                self.c = _sentinel

    def _next_not_eq(self, excludes: Container[str]):
        if self.c in excludes:
            for c in self._iter_text:
                if c not in excludes:
                    self.c = c
                    break
            else:
                self.c = _sentinel

    def _decode_table(self):
        table: dict[Any, Any] | list[Any] = {}
        idx = 0

        self._next()
        while True:
            self._next_not_eq(WHITESPACE)

            if self.c == '}':

                # Convert table to list if k(0) = 1 and k = k(n-1) + 1, ...
                if (
                    table
                    and all(map(eq, table, count(1)))
                    # bool is a subclass of int in Python but not in Lua
                    and not any(isinstance(k, bool) for k in islice(table, 0, 2))
                ):
                    table = list(table.values())

                self._next()
                return table

            elif self.c == ',':

                self._next()

            else:

                is_val_long_string_literal = False

                if self.c == '[':
                    self._next()
                    if self.c == '[':
                        is_val_long_string_literal = True

                item = self.decode()
                self._next_not_eq(WHITESPACE_OR_CLOSING_SQ_BR)

                c = self.c
                if c and c in '=,':
                    self._next()

                    if c == '=':
                        if is_val_long_string_literal:
                            raise ParseError('malformed key', item)

                        # nil key produces a runtime error in Lua
                        if item is None:
                            raise ParseError('table keys cannot be nil')

                        # Item is a key
                        value = self.decode()
                        if (
                            # nil values are not persisted in Lua tables
                            value is not None
                            # Where the key is a valid index key-less values take precedence
                            and (not isinstance(item, int) or isinstance(item, bool) or item > idx)
                        ):
                            table[item] = value
                        continue

                if item is not None:
                    idx += 1
                    table[idx] = item

    def _decode_string(self):
        s = ''
        start = self.c
        end = None
        prev_was_slash = False

        if start == '[':
            self._next_not_eq('[')
            s += self.c
            end = ']'
        else:
            end = start

        for c in self._iter_text:
            if prev_was_slash:
                prev_was_slash = False

                if c != end:
                    s += '\\'
            elif c == end:
                break
            elif c == '\\' and start == end:
                prev_was_slash = True
                continue

            s += c

        self._next()
        if start != end:
            # Strip multiple closing brackets
            self._next_not_eq(end)

        return s

    def _decode_bare_word(self):
        s = self.c
        for c in self._iter_text:
            new_s = s + c
            if match_bare_word.match(new_s):
                s = new_s
            else:
                break

        self._next()

        if s == "true":
            return True
        elif s == "false":
            return False
        elif s == "nil":
            return None
        return s

    def _decode_number(self):
        def get_digits():
            n = ''

            for c in self._iter_text:
                if c in DIGITS:
                    n += c
                else:
                    self.c = c
                    break
            else:
                self.c = _sentinel

            return n

        n = ''

        if self.c == '-':
            c = self.c
            self._next()
            if self.c == '-':

                # This is a comment - skip to the end of the line
                self._next_eq(NEWLINE)
                return None

            elif not self.c or self.c not in DIGITS:
                raise ParseError('malformed number (no digits after minus sign)', c + self.c)

            n += c

        n += self.c + get_digits()
        if n == '0' and self.c in HEXDELIMS:

            n += self.c

            for c in self._iter_text:
                if c in HEXDIGITS:
                    n += c
                else:
                    self.c = c
                    break
            else:
                self.c = _sentinel

        else:

            if self.c == '.':
                n += self.c + get_digits()

            if self.c in EXPONENTS:
                n += self.c
                self._next()  # +-
                n += self.c + get_digits()

        try:
            return int(n, 0)
        except ValueError:
            return float(n)

    def decode(self):
        self._next_not_eq(WHITESPACE)
        if not self.c:
            raise ParseError('input is empty')
        elif self.c == '{':
            return self._decode_table()
        elif self.c in '\'"[':
            return self._decode_string()
        elif self.c == '-' or self.c in DIGITS:
            return self._decode_number()
        else:
            return self._decode_bare_word()


def loads(s: str) -> Any:
    return _SLPP(s).decode()



"""
Add dependency table.
"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '7204944522b1'
down_revision = '2bfcfe824fe0'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'pkg_dep',
        sa.Column('_id', sa.Integer(), nullable=False),
        sa.Column('id', sa.String(), nullable=False),
        sa.Column('pkg_origin', sa.String(), nullable=False),
        sa.Column('pkg_id', sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ['pkg_origin', 'pkg_id'],
            ['pkg.origin', 'pkg.id'],
        ),
        sa.PrimaryKeyConstraint('_id'),
    )


def downgrade():
    op.drop_table('pkg_dep')



from __future__ import annotations

import asyncio
from collections.abc import Callable as C
from functools import partial
import json
import shutil
from textwrap import dedent
from unittest import mock

from click.testing import CliRunner, Result
from prompt_toolkit.application import create_app_session
from prompt_toolkit.input import create_pipe_input
from prompt_toolkit.output import DummyOutput
import pytest

from instawow import __version__
from instawow.cli import cli
from instawow.common import Flavour
from instawow.config import Config
from instawow.models import PkgList


@pytest.fixture
def feed_pt():
    pipe_input = create_pipe_input()
    with create_app_session(input=pipe_input, output=DummyOutput()):
        yield pipe_input.send_text
    pipe_input.close()


@pytest.fixture
def run(
    monkeypatch: pytest.MonkeyPatch,
    event_loop: asyncio.AbstractEventLoop,
    iw_config: Config,
):
    monkeypatch.setattr('prompt_toolkit.shortcuts.progress_bar.ProgressBar', mock.MagicMock())
    monkeypatch.setattr('asyncio.run', event_loop.run_until_complete)
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', str(iw_config.global_config.config_dir))
    yield partial(CliRunner().invoke, cli, catch_exceptions=False)


@pytest.fixture
def install_molinari_and_run(
    run: C[[str], Result],
):
    run('install curse:molinari')
    yield run


@pytest.fixture
def pretend_install_molinari_and_run(
    iw_config: Config,
    run: C[[str], Result],
):
    molinari = iw_config.addon_dir / 'Molinari'
    molinari.mkdir()
    (molinari / 'Molinari.toc').write_text(
        '''\
## X-Curse-Project-ID: 20338
## X-WoWI-ID: 13188
'''
    )
    yield run


@pytest.fixture
def molinari_version_suffix(
    iw_config: Config,
):
    if iw_config.game_flavour is Flavour.vanilla_classic:
        return '-classic'
    if iw_config.game_flavour is Flavour.burning_crusade_classic:
        return '-bcc'
    else:
        return ''


@pytest.mark.parametrize(
    'alias', ['curse:molinari', 'wowi:13188-molinari', 'tukui:1', 'github:p3lim-wow/Molinari']
)
def test_valid_pkg_lifecycle(
    run: C[[str], Result],
    alias: str,
):
    assert run(f'install {alias}').output.startswith(f'✓ {alias}\n  installed')
    assert run(f'install {alias}').output == f'✗ {alias}\n  package already installed\n'
    assert run(f'update {alias}').output == f'✗ {alias}\n  package is up to date\n'
    assert run(f'remove {alias}').output == f'✓ {alias}\n  removed\n'
    assert run(f'update {alias}').output == f'✗ {alias}\n  package is not installed\n'
    assert run(f'remove {alias}').output == f'✗ {alias}\n  package is not installed\n'


@pytest.mark.parametrize('alias', ['curse:gargantuan-wigs'])
def test_nonexistent_pkg_lifecycle(
    run: C[[str], Result],
    alias: str,
):
    assert run(f'install {alias}').output == f'✗ {alias}\n  package does not exist\n'
    assert run(f'update {alias}').output == f'✗ {alias}\n  package is not installed\n'
    assert run(f'remove {alias}').output == f'✗ {alias}\n  package is not installed\n'


@pytest.mark.parametrize('alias', ['foo:bar'])
def test_invalid_source_lifecycle(
    run: C[[str], Result],
    alias: str,
):
    assert run(f'install {alias}').output == f'✗ {alias}\n  package source is invalid\n'
    assert run(f'update {alias}').output == f'✗ {alias}\n  package is not installed\n'
    assert run(f'remove {alias}').output == f'✗ {alias}\n  package is not installed\n'


def test_reconciled_folder_conflict_on_install(
    run: C[[str], Result],
):
    assert run('install curse:molinari').output.startswith('✓ curse:molinari\n  installed')
    assert run('install wowi:13188-molinari').output == (
        '✗ wowi:13188-molinari\n'
        '  package folders conflict with installed package Molinari\n'
        '    (curse:20338)\n'
    )


def test_unreconciled_folder_conflict_on_install(
    iw_config: Config,
    run: C[[str], Result],
):
    iw_config.addon_dir.joinpath('Molinari').mkdir()
    assert (
        run('install curse:molinari').output
        == "✗ curse:molinari\n  package folders conflict with 'Molinari'\n"
    )
    assert run('install --replace curse:molinari').output.startswith(
        '✓ curse:molinari\n  installed'
    )


def test_keep_folders_on_remove(
    iw_config: Config,
    install_molinari_and_run: C[[str], Result],
):
    assert (
        install_molinari_and_run('remove --keep-folders curse:molinari').output
        == '✓ curse:molinari\n  removed\n'
    )
    assert iw_config.addon_dir.joinpath('Molinari').is_dir()


def test_version_strategy_lifecycle(
    run: C[[str], Result],
):
    assert run('install curse:molinari').output.startswith(
        '✓ curse:molinari\n  installed 90105.81-Release'
    )
    assert (
        run('install --version foo curse:molinari').output
        == '✗ curse:molinari\n  package already installed\n'
    )
    assert run('update curse:molinari').output == '✗ curse:molinari\n  package is up to date\n'
    assert run('remove curse:molinari').output == '✓ curse:molinari\n  removed\n'
    assert (
        run('install --version foo curse:molinari').output
        == f"✗ curse:molinari\n  version foo not found\n"
    )
    assert (
        run('install --version 80000.57-Release curse:molinari').output
        == '✓ curse:molinari\n  installed 80000.57-Release\n'
    )
    assert run('update').output == '✗ curse:molinari\n  package is pinned\n'
    assert run('update curse:molinari').output == '✗ curse:molinari\n  package is pinned\n'
    assert run('remove curse:molinari').output == '✓ curse:molinari\n  removed\n'


def test_install_options(
    run: C[[str], Result],
    molinari_version_suffix: str,
):
    assert run(
        'install'
        ' curse:molinari'
        ' -s latest curse:molinari'
        ' --version 80000.57-Release curse:molinari'
    ).output == dedent(
        f'''\
        ✓ curse:molinari
          installed 90105.81-Release{molinari_version_suffix}
        ✗ curse:molinari
          package folders conflict with installed package Molinari
            (curse:20338)
        ✗ curse:molinari
          package folders conflict with installed package Molinari
            (curse:20338)
        '''
    )


def test_install_option_order_is_respected(
    run: C[[str], Result],
):
    assert run(
        'install'
        ' --version 80000.57-Release curse:molinari'
        ' -s latest curse:molinari'
        ' curse:molinari'
    ).output == dedent(
        '''\
        ✓ curse:molinari
          installed 80000.57-Release
        ✗ curse:molinari
          package folders conflict with installed package Molinari
            (curse:20338)
        ✗ curse:molinari
          package folders conflict with installed package Molinari
            (curse:20338)
        '''
    )


def test_install_argument_is_not_required(
    run: C[[str], Result],
    molinari_version_suffix: str,
):
    assert run(
        'install -s latest curse:molinari --version 80000.57-Release curse:molinari'
    ).output == dedent(
        f'''\
        ✓ curse:molinari
          installed 90105.81-Release{molinari_version_suffix}
        ✗ curse:molinari
          package folders conflict with installed package Molinari
            (curse:20338)
        '''
    )


def test_configure__show_active_profile(
    iw_config: Config,
    run: C[[str], Result],
):
    assert run('configure --show-active').output == iw_config.json(indent=2) + '\n'


def test_configure__create_new_profile(
    feed_pt: C[[str], None],
    iw_config: Config,
    run: C[[str], Result],
):
    feed_pt(f'{iw_config.addon_dir}\r\rY\r')
    assert run('-p foo configure').output == (
        'Navigate to https://github.com/login/device and paste the code below:\n'
        '  WDJB-MJHT\n'
        'Waiting...\n'
        'Configuration written to:\n'
        f'  {iw_config.global_config.config_dir / "config.json"}\n'
        f'  {iw_config.global_config.config_dir / "profiles/foo/config.json"}\n'
    )


def test_configure__update_existing_profile_with_opts(
    feed_pt: C[[str], None],
    iw_config: Config,
    run: C[[str], Result],
):
    feed_pt(f'Y\r')
    assert run('configure auto_update_check').output == (
        'Configuration written to:\n'
        f'  {iw_config.global_config.config_dir / "config.json"}\n'
        f'  {iw_config.global_config.config_dir / "profiles/__default__/config.json"}\n'
    )


@pytest.mark.parametrize('options', ['', '--undo'])
def test_rollback__pkg_not_installed(
    run: C[[str], Result],
    options: str,
):
    assert (
        run(f'rollback {options} curse:molinari').output
        == '✗ curse:molinari\n  package is not installed\n'
    )


@pytest.mark.parametrize('options', ['', '--undo'])
def test_rollback__unsupported(
    run: C[[str], Result],
    options: str,
):
    assert run('install wowi:13188-molinari').exit_code == 0
    assert (
        run(f'rollback {options} wowi:13188-molinari').output
        == "✗ wowi:13188-molinari\n  'version' strategy is not valid for source\n"
    )


def test_rollback__single_version(
    run: C[[str], Result],
):
    assert run('install curse:molinari').exit_code == 0
    assert (
        run('rollback curse:molinari').output == '✗ curse:molinari\n  cannot find older versions\n'
    )


def test_rollback__multiple_versions(
    feed_pt: C[[str], None],
    run: C[[str], Result],
    molinari_version_suffix: str,
):
    assert run('install --version 80000.57-Release curse:molinari').exit_code == 0
    assert run('remove curse:molinari').exit_code == 0
    assert run('install curse:molinari').exit_code == 0
    feed_pt('\r\r')
    assert run('rollback curse:molinari').output == dedent(
        f'''\
        ✓ curse:molinari
          updated 90105.81-Release{molinari_version_suffix} to 80000.57-Release
        '''
    )


def test_rollback__multiple_versions_promptless(
    run: C[[str], Result],
    molinari_version_suffix: str,
):
    assert run('install --version 80000.57-Release curse:molinari').exit_code == 0
    assert run('remove curse:molinari').exit_code == 0
    assert run('install curse:molinari').exit_code == 0
    assert run('rollback --version 80000.57-Release curse:molinari').output == dedent(
        f'''\
        ✓ curse:molinari
          updated 90105.81-Release{molinari_version_suffix} to 80000.57-Release
        '''
    )


@pytest.mark.parametrize('options', ['', '--undo'])
def test_rollback__rollback_multiple_versions(
    feed_pt: C[[str], None],
    run: C[[str], Result],
    molinari_version_suffix: str,
    options: str,
):
    assert run('install curse:molinari').exit_code == 0
    assert run('remove curse:molinari').exit_code == 0
    assert run('install --version 80000.57-Release curse:molinari').exit_code == 0
    feed_pt('\r\r')
    assert run(f'rollback {options} curse:molinari').output == dedent(
        f'''\
        ✓ curse:molinari
          updated 80000.57-Release to 90105.81-Release{molinari_version_suffix}
        '''
    )


def test_rollback__cannot_use_version_with_undo(
    run: C[[str], Result],
):
    result = run('rollback --version foo --undo curse:molinari')
    assert result.exit_code == 2
    assert 'Cannot use "--version" with "--undo"' in result.output


def test_reconcile__list_unreconciled(
    pretend_install_molinari_and_run: C[[str], Result],
):
    assert pretend_install_molinari_and_run('reconcile --list-unreconciled').output == (
        # fmt: off
        'unreconciled\n'
        '------------\n'
        'Molinari    \n'
        # fmt: on
    )


def test_reconcile__auto_reconcile(
    pretend_install_molinari_and_run: C[[str], Result],
    molinari_version_suffix: str,
):
    assert pretend_install_molinari_and_run('reconcile --auto').output == dedent(
        f'''\
        ✓ curse:molinari
          installed 90105.81-Release{molinari_version_suffix}
        '''
    )


def test_reconcile__abort_interactive_reconciliation(
    feed_pt: C[[str], None],
    pretend_install_molinari_and_run: C[[str], Result],
):
    feed_pt('\x03')  # ^C
    assert pretend_install_molinari_and_run('reconcile').output.endswith('Aborted!\n')


def test_reconcile__complete_interactive_reconciliation(
    feed_pt: C[[str], None],
    pretend_install_molinari_and_run: C[[str], Result],
    molinari_version_suffix,
):
    feed_pt('\r\r')
    assert pretend_install_molinari_and_run('reconcile').output.endswith(
        dedent(
            f'''\
            ✓ curse:molinari
              installed 90105.81-Release{molinari_version_suffix}
            '''
        )
    )


def test_reconcile__reconciliation_complete(
    run: C[[str], Result],
):
    assert run('reconcile').output == 'No add-ons left to reconcile.\n'


def test_reconcile__rereconcile(
    feed_pt: C[[str], None],
    install_molinari_and_run: C[[str], Result],
):
    feed_pt('\r\r')
    assert install_molinari_and_run('reconcile --installed').output == dedent(
        f'''\
        ✓ curse:molinari
          removed
        ✓ github:p3lim-wow/molinari
          installed 90105.81-Release
        '''
    )


def test_reconcile__cannot_use_auto_with_installed(
    run: C[[str], Result],
):
    result = run('reconcile --auto --installed')
    assert result.exit_code == 2
    assert 'Cannot use "--auto" with "--installed"' in result.output


@pytest.mark.skip
def test_search__no_results(
    run: C[[str], Result],
):
    assert run('search ∅').output == 'No results found.\n'


def test_search__exit_without_selecting(
    feed_pt: C[[str], None],
    run: C[[str], Result],
):
    feed_pt('\r')  # enter
    assert run('search molinari').output == ''


def test_search__exit_after_selection(
    feed_pt: C[[str], None],
    run: C[[str], Result],
):
    feed_pt(' \rn')  # space, enter, "n"
    assert run('search molinari').output == ''


def test_search__install_one(
    feed_pt: C[[str], None],
    run: C[[str], Result],
    molinari_version_suffix: str,
):
    feed_pt(' \r\r')  # space, enter, enter
    assert run('search molinari --source curse').output == dedent(
        f'''\
        ✓ curse:molinari
          installed 90105.81-Release{molinari_version_suffix}
        '''
    )


def test_search__install_multiple_conflicting(
    feed_pt: C[[str], None],
    run: C[[str], Result],
):
    feed_pt(' \x1b[B \r\r')  # space, arrow down, space, enter, enter
    assert run('search molinari').output == dedent(
        f'''\
        ✓ wowi:13188-molinari
          installed 90105.81-Release
        ✗ curse:molinari
          package folders conflict with installed package Molinari
            (wowi:13188)
        '''
    )


def test_changelog_output(
    install_molinari_and_run: C[[str], Result],
):
    output = (
        'Changes in 90105.81-Release:'
        if shutil.which('pandoc')
        else '<h3>Changes in 90105.81-Release:</h3>'
    )
    assert install_molinari_and_run('view-changelog curse:molinari').output.startswith(output)


def test_changelog_output_no_convert(
    install_molinari_and_run: C[[str], Result],
):
    assert install_molinari_and_run(
        'view-changelog --no-convert curse:molinari'
    ).output.startswith('<h3>Changes in 90105.81-Release:</h3>')


def test_argless_changelog_output(
    install_molinari_and_run: C[[str], Result],
):
    output = (
        'curse:molinari:\n  Changes in 90105.81-Release:'
        if shutil.which('pandoc')
        else 'curse:molinari:\n  <h3>Changes in 90105.81-Release:</h3>'
    )
    assert install_molinari_and_run('view-changelog').output.startswith(output)


def test_argless_changelog_output_no_convert(
    install_molinari_and_run: C[[str], Result],
):
    assert install_molinari_and_run('view-changelog --no-convert').output.startswith(
        'curse:molinari:\n  <h3>Changes in 90105.81-Release:</h3>'
    )


@pytest.mark.parametrize(
    'command, exit_code',
    [
        ('list mol', 0),
        ('info foo', 0),
        ('reveal mol', 0),
        ('reveal foo', 1),
    ],
)
def test_exit_codes_with_substr_match(
    monkeypatch: pytest.MonkeyPatch,
    install_molinari_and_run: C[[str], Result],
    command: str,
    exit_code: int,
):
    monkeypatch.setattr('click.launch', lambda *_, **__: ...)
    assert install_molinari_and_run(command).exit_code == exit_code


def test_can_list_with_substr_match(
    install_molinari_and_run: C[[str], Result],
):
    assert install_molinari_and_run('list mol').output == 'curse:molinari\n'
    assert install_molinari_and_run('list foo').output == ''
    assert install_molinari_and_run('list -f detailed mol').output.startswith('curse:molinari')
    (molinari,) = json.loads(install_molinari_and_run('list -f json mol').output)
    assert (molinari['source'], molinari['slug']) == ('curse', 'molinari')


def test_json_export(
    install_molinari_and_run: C[[str], Result],
):
    output = install_molinari_and_run('list -f json').output
    assert PkgList.parse_raw(output).__root__[0].name == 'Molinari'


def test_show_version(
    run: C[[str], Result],
):
    assert run('--version').output == f'instawow, version {__version__}\n'


def test_plugin_hook_command_can_be_invoked(
    run: C[[str], Result],
):
    pytest.importorskip('instawow_test_plugin')
    assert run('foo').output == 'success!\n'
